"""Helpers for writing lark transformers."""
from typing import Any, Optional, Tuple, TypeVar, Union

import lark

from .common import FullLoadContext, LoadContext


def context_from_token(fn: str, token: lark.Token) -> FullLoadContext:
    """Get a LoadContext from a lark Token."""
    return (LoadContext(name=fn, line=token.line), )


def context_from_tree(fn: str, tree: lark.Tree) -> FullLoadContext:
    """Get a LoadContext from a lark Tree."""
    return (LoadContext(name=fn, line=tree.meta.line), )


T = TypeVar("T")


@staticmethod
def tuple_args(*objects: T) -> Tuple[T, ...]:
    """Transformer helper to get back the *args tuple."""
    return objects


@staticmethod
def listify(*objects: Any):
    """Transformer helper to listify *args."""
    return list(objects)


@staticmethod
def listify_strings(*objects: Union[str, lark.Token]):
    """Transformer helper to listify *args and stringify each arg."""
    return list(str(obj) for obj in objects)


@staticmethod
def stringify(obj: lark.Token) -> str:
    """Transformer helper to stringify a single argument."""
    return str(obj)


@staticmethod
def pass_through(obj: Optional[T] = None) -> Optional[T]:
    """Transformer helper to pass through an optional single argument."""
    return obj


@staticmethod
def dictify(*tuples: Tuple[Any, Any]) -> dict:
    """Transformer helper to stringify a single argument."""
    return dict(tuples)


@staticmethod
def ignore(*args: Any) -> None:
    """Transformer helper to drop the subtree."""



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, Optional, Union

from .common import (AsynPortBase, FullLoadContext, RecordField,
                     RecordInstance, ShellStateHandler)

logger = logging.getLogger(__name__)


@dataclass
class AsynPort(AsynPortBase):
    context: FullLoadContext
    name: str
    options: Dict[str, AsynPortOption] = field(default_factory=dict)
    metadata: Dict[str, str] = field(default_factory=dict)
    motors: Dict[str, AsynMotor] = field(default_factory=dict)

    _jinja_format_: ClassVar[dict] = {
        "console": """\
{{obj|classname}}:
    name: {{name}}
    options: {{options}}
    metadata: {{metadata}}
    Defined:
{% for ctx in context %}
        * {{ctx}}
{% endfor %}
{% for motor in motors %}
{% set motor_text = render_object(motor, "console") %}
        {{ motor_text | indent(8)}}
{% endfor %}
""",

    }


@dataclass
class AsynMotor(AsynPortBase):
    context: FullLoadContext
    name: str
    metadata: Dict[str, Union[str, int, float]] = field(default_factory=dict)
    parent: Optional[str] = None


@dataclass
class AsynIPPort(AsynPort):
    hostInfo: str = ""
    priority: str = ""
    noAutoConnect: str = ""
    noProcessEos: str = ""


@dataclass
class AsynSerialPort(AsynPort):
    ttyName: str = ""
    priority: str = ""
    noAutoConnect: str = ""
    noProcessEos: str = ""


@dataclass
class AdsAsynPort(AsynPort):
    ipaddr: str = ""
    amsaddr: str = ""
    amsport: int = 0
    asynParamTableSize: int = 0
    priority: int = 0
    noAutoConnect: int = 0
    defaultSampleTimeMS: int = 0
    maxDelayTimeMS: int = 0
    adsTimeoutMS: int = 0
    defaultTimeSource: str = ""

    _jinja_format_: ClassVar[dict] = {
        "console": AsynPort._jinja_format_["console"] + """
    ipaddr: {{ipaddr}}
    amsaddr: {{amsaddr}}
    amsport: {{amsport}}
    asynParamTableSize: {{asynParamTableSize}}
    priority: {{priority}}
    noAutoConnect: {{noAutoConnect}}
    defaultSampleTimeMS: {{defaultSampleTimeMS}}
    maxDelayTimeMS: {{maxDelayTimeMS}}
    adsTimeoutMS: {{adsTimeoutMS}}
    defaultTimeSource: {{defaultTimeSource}}
"""
    }


@dataclass
class AsynPortMultiDevice(AsynPortBase):
    context: FullLoadContext
    name: str
    metadata: dict = field(default_factory=dict)
    motors: Dict[str, AsynMotor] = field(default_factory=dict)
    devices: Dict[str, AsynPortDevice] = field(default_factory=dict)


@dataclass
class AsynPortDevice(AsynPortBase):
    context: FullLoadContext
    name: str = ""
    options: Dict[str, AsynPortOption] = field(default_factory=dict)
    metadata: Dict[str, str] = field(default_factory=dict)
    motors: Dict[str, AsynMotor] = field(default_factory=dict)


@dataclass
class AsynPortOption:
    context: FullLoadContext
    key: str
    value: str


_handler = ShellStateHandler.generic_handler_decorator


@dataclass
class AsynState(ShellStateHandler):
    """
    Asyn IOC shell state handler / container.

    Contains hooks for asyn-related commands and state information.

    Attributes
    ----------
    ports : Dict[str, asyn.AsynPortBase]
        Asyn ports defined by name.
    """
    metadata_key: ClassVar[str] = "asyn"
    ports: Dict[str, AsynPortBase] = field(default_factory=dict)

    @_handler
    def handle_drvAsynSerialPortConfigure(
        self,
        portName: str,
        ttyName: str,
        priority: int = 0,
        noAutoConnect: int = 0,
        noProcessEos: int = 0,
    ):
        # SLAC-specific, but doesn't hurt anyone
        self.ports[portName] = AsynSerialPort(
            context=self.get_load_context(),
            name=portName,
            ttyName=ttyName,
            priority=priority,
            noAutoConnect=noAutoConnect,
            noProcessEos=noProcessEos,
        )

    @_handler
    def handle_drvAsynIPPortConfigure(
        self,
        portName: str,
        hostInfo: str,
        priority: int = 0,
        noAutoConnect: int = 0,
        noProcessEos: int = 0,
    ):
        # SLAC-specific, but doesn't hurt anyone
        self.ports[portName] = AsynIPPort(
            context=self.get_load_context(),
            name=portName,
            hostInfo=hostInfo,
            priority=priority,
            noAutoConnect=noAutoConnect,
            noProcessEos=noProcessEos,
        )

    @_handler
    def handle_asynSetOption(self, name: str, addr: str, key: str, value: str):
        port = self.ports[name]
        opt = AsynPortOption(
            context=self.get_load_context(),
            key=key,
            value=value,
        )

        if isinstance(port, AsynPortMultiDevice):
            port.devices[addr].options[key] = opt
        else:
            port.options[key] = opt

    def get_port_from_record(self, inst: RecordInstance) -> Optional[AsynPort]:
        """Given a record, return its related asyn port."""
        if inst.is_pva:
            return

        rec_field: Optional[RecordField]
        rec_field = inst.fields.get("INP", inst.fields.get("OUT", None))
        if rec_field is None:
            return

        if not isinstance(rec_field.value, str):
            # No PVAccess links just yet
            return

        value = rec_field.value.strip()
        if value.startswith("@asyn"):
            try:
                asyn_args = value.split("@asyn")[1].strip(" ()")
                asyn_port, *_ = asyn_args.split(",")
                return self.ports.get(asyn_port.strip(), None)
            except Exception:
                logger.debug("Failed to parse asyn string", exc_info=True)

    def annotate_record(self, record: RecordInstance) -> Optional[Dict[str, Any]]:
        port = self.get_port_from_record(record)
        if port is not None:
            parent_port = getattr(port, "parent", None)
            ports = [
                port_ for port_ in [parent_port, port]
                if port_ is not None
            ]
            return {"ports": ports}



from typing import Optional

from epicsmacrolib import IocshRedirect, IocshSplit, split_iocsh_line

from .common import IocshResult, LoadContext
from .macro import MacroContext

__all__ = ["split_iocsh_line", "parse_iocsh_line", "IocshRedirect", "IocshSplit"]


def parse_iocsh_line(
    line: str, *,
    context: Optional[LoadContext] = None,
    prompt: str = "epics>",
    macro_context: Optional[MacroContext] = None,
    string_encoding: str = "latin-1",
) -> IocshResult:
    """
    Parse an IOC shell line into an IocshResult.

    Parameters
    ----------
    line : str
       The line to parse.

    context : LoadContext, optional
       The load context to populate the result with.

    prompt : str, optional
       Replicating the EPICS source code, specify the state of the prompt
       here.  Defaults to "epics>".  If unset as in prior to IOC init,
       lines that do not start with "#-" will be eched.

    Returns
    -------
    IocshResult
       A partially filled IocshResult, ready for interpreting by a
       higher-level function.
    """
    result = IocshResult(context=context, line=line)
    # Skip leading whitespace
    line = line.lstrip()

    if not line.startswith("#-"):
        result.outputs.append(line)

    if line.startswith('#'):
        # Echo non-empty lines read from a script.
        # Comments delineated with '#-' aren't echoed.
        return result

    if macro_context is not None:
        line = macro_context.expand(line)

    # * Skip leading white-space coming from a macro
    line = line.lstrip()

    # * Echo non-empty lines read from a script.
    # * Comments delineated with '#-' aren't echoed.
    if not prompt:
        if not line.startswith('#-'):
            result.outputs.append(line)

    # * Ignore lines that became a comment or empty after macro expansion
    if not line or line.startswith('#'):
        return result

    split = split_iocsh_line(line, string_encoding=string_encoding)
    result.argv = split.argv

    # Only set the following if necessary; apischema can skip serialization
    # otherwise.
    if split.redirects:
        result.redirects = list(split.redirects.values())

    if split.error:
        result.error = split.error

    return result



import shutil
import sys

import pytest

from ..common import IocMetadata

SUPPORTED_PLATFORM = sys.platform in {"linux"}


platform_skip = pytest.mark.skipif(
    not SUPPORTED_PLATFORM,
    reason=f"Not supported on {sys.platform}",
)

gdb_unavailable_skip = pytest.mark.skipif(
    shutil.which("gdb") is None,
    reason="GDB unavailable"
)


softioc = shutil.which("softIoc")

softioc_unavailable_skip = pytest.mark.skipif(
    softioc is None,
    reason="softIoc unavailable"
)


@pytest.mark.xfail(reason="Makes too many SLAC/LCLS/PCDS/ECS assumptions.")
@pytest.mark.asyncio
@platform_skip
@gdb_unavailable_skip
@softioc_unavailable_skip
async def test_gdb_info():
    md = IocMetadata(binary=str(softioc))

    info = await md.get_binary_information()
    assert info is not None, "Script appears to have failed"
    print("Got information", info)

    assert len(md.variables)
    assert len(md.commands)
    assert md.variables == info.variables
    assert md.commands == info.commands
    assert md.base_version == info.base_version
    assert any(
        (
            "3." in md.base_version,
            "7." in md.base_version,
        )
    )
    assert "dbLoadRecords" in md.commands
    assert "CASDEBUG" in md.variables



import ast
import collections
import copy
import dataclasses
import html
import logging
import textwrap
from dataclasses import dataclass
from typing import Any, ClassVar, Dict, Generator, List, Optional, Tuple, Union

import graphviz as gv

from .common import (FullLoadContext, LoadContext, PVRelations,
                     ScriptPVRelations)
from .db import Database, RecordField, RecordInstance, RecordType
from .gv_compat import AsyncDigraph

logger = logging.getLogger(__name__)

# TODO: refactor this to not be graphviz-dependent; instead return node/link
# information in terms of dataclasses


@dataclass()
class GraphNode:
    #: The integer ID of the node
    id: str
    #: The node label
    label: str
    #: The text to show in the node
    text: str
    #: Options to pass to graphviz.
    options: dict = dataclasses.field(default_factory=dict)
    #: Highlight the node in the graph?
    highlighted: bool = False
    #: User-specified metadata
    metadata: Dict[str, Any] = dataclasses.field(default_factory=dict)

    def add_text_line(self, line: str, delimiter: str = "\n", only_unique: bool = True):
        """
        Add a line of text to the node.

        Parameters
        ----------
        line : str
            The line to add.

        delimiter : str, optional
            The between-line delimiter.

        only_unique : bool, optional
            Only add unique lines to the text.
        """
        if not self.text.strip():
            self.text = line
        elif not only_unique or line not in self.text:
            self.text = delimiter.join((self.text, line))

    def __hash__(self):
        return hash(self.id)


@dataclass
class GraphEdge:
    #: The source node.
    source: GraphNode
    #: The destination node.
    destination: GraphNode
    #: The source node.
    source_port: Optional[str] = None
    #: The destination node.
    destination_port: Optional[str] = None
    #: Options to pass to graphviz.
    options: dict = dataclasses.field(default_factory=dict)

    def __hash__(self) -> int:
        return hash(
            (self.source.id, self.destination.id, self.source_port, self.destination_port)
        )

    @property
    def source_with_port(self) -> str:
        """Source name including the port name, for graphviz."""
        if self.source_port is None:
            return self.source.id
        return f"{self.source.id}:{self.source_port}_value:e"

    @property
    def destination_with_port(self) -> str:
        """Destination name including the port name, for graphviz."""
        if self.destination_port is None:
            return self.destination.id
        return f"{self.destination.id}:{self.destination_port}_name:w"


class _GraphHelper:
    """
    A base class for helping build graphviz digraphs.
    """

    shapes: Tuple[str, str] = ("rectangle", "box3d")
    fill_colors: Tuple[str, str] = ("white", "bisque")
    newline: str = '<br align="left"/>'
    nodes: Dict[str, GraphNode]
    edges: List[GraphEdge]

    def __init__(self):
        self._node_id = 0
        self.nodes = {}
        self.edges = []

    @property
    def edge_pairs(self) -> Generator[Tuple[GraphNode, GraphNode], None, None]:
        for edge in self.edges:
            yield (edge.source, edge.destination)

    def get_node(
        self, label: str, text: Optional[str] = None
    ) -> GraphNode:
        """Create a new node in the graph."""
        if label not in self.nodes:
            self._node_id += 1
            self.nodes[label] = GraphNode(
                id=str(self._node_id),
                text=text or label,
                label=label
            )
            logger.debug("Created node %s", label)

        if text and text.strip():
            self.nodes[label].add_text_line(text)

        return self.nodes[label]

    def add_edge(
        self,
        source: str,
        destination: str,
        allow_dupes: bool = False,
        source_port: Optional[str] = None,
        dest_port: Optional[str] = None,
        **options
    ) -> GraphEdge:
        """Create a new edge in the graph."""
        edge = GraphEdge(
            self.get_node(source),
            self.get_node(destination),
            source_port=source_port,
            destination_port=dest_port,
            options=options
        )
        if edge in self.edges and not allow_dupes:
            return self.edges[self.edges.index(edge)]
        self.edges.append(edge)
        return edge

    def _ready_for_digraph(self, graph: gv.Digraph):
        """Hook when the user calls ``to_digraph``."""
        raise NotImplementedError()

    def to_digraph(
        self,
        graph: Optional[gv.Digraph] = None,
        engine: str = "dot",
        font_name: Optional[str] = "Courier",
        format: str = "pdf",
    ) -> gv.Digraph:
        """
        Create a graphviz digraph.

        Parameters
        ----------
        graph : graphviz.Graph, optional
            Graph instance to use.  New one created if not specified.
        engine : str, optional
            Graphviz engine (dot, fdp, etc).
        font_name : str, optional
            Font name to use for all nodes and edges.
        format :
            The output format used for rendering (``'pdf'``, ``'png'``, ...).
        """
        graph = graph or AsyncDigraph(format=format)

        # Call the subclass hook:
        self._ready_for_digraph(graph)

        if engine is not None:
            graph.engine = engine

        if font_name is not None:
            graph.attr("graph", fontname=font_name)
            graph.attr("node", fontname=font_name)
            graph.attr("edge", fontname=font_name)

        for node in self.nodes.values():
            text = self.newline.join(node.text.splitlines())
            graph.node(
                node.id,
                label=f"< {text} >",
                shape=self.shapes[node.highlighted],
                fillcolor=self.fill_colors[node.highlighted],
                style="filled",
                **node.options
            )

        # add all of the edges between graphs
        for edge in self.edges:
            graph.edge(
                edge.source_with_port,
                edge.destination_with_port,
                **edge.options
            )
        return graph

    @staticmethod
    def clean_code(obj):
        """Clean C-like code for graphviz."""
        code = str(obj or "").strip("{}")
        return html.escape(textwrap.dedent(code).strip(" \n"))


@dataclass
class LinkInfo:
    record1: RecordInstance
    field1: RecordField
    record2: RecordInstance
    field2: RecordField
    info: List[str]


def is_supported_link(link: str) -> bool:
    if link.startswith("#") or link.startswith("0x") or link.startswith("@"):
        return False
    try:
        ast.literal_eval(link)
    except Exception:
        # Should not be an integer literal
        return True

    return False


def _get_links_for_record(
    record: RecordInstance, record_types: Optional[Dict[str, RecordType]] = None
) -> Generator[Tuple[RecordField, str, List[str]], None, None]:
    """
    Get links for the provided record, referring back to the record_types dict if necessary.
    """
    if record.has_dbd_info:
        yield from record.get_links()
    elif record_types is not None:
        rec1_rtype = record_types.get(record.record_type, None)
        if rec1_rtype is not None:
            yield from rec1_rtype.get_links_for_record(record)


_unset_ctx: FullLoadContext = (LoadContext("unknown", 0), )


def _field_from_record_relation(
    record: Optional[RecordInstance],
    field_name: str,
    link_text: str,
    record_types: Dict[str, RecordType]
) -> Optional[RecordField]:
    """
    Create a RecordField instance based on what we know, given the parameters.

    Parameters
    ----------
    record : RecordInstance or None
        An (optional) record instance, if available in the database.

    field_name : str
        A field name for the record (field_name).

    link_text : str, optional
        The link text - likely the record name - referring to ``record``.

    record_types : dict, optional
        Record type information from a database definition.
    """
    if record is None:
        # Case 1: The linked record is *not* in the database.

        if not is_supported_link(link_text):
            return None

        return RecordField(
            dtype="unknown",
            name=field_name,
            value="(unknown-record)",
            context=_unset_ctx,
        )

    if field_name in record.fields:
        # Case 2: The linked record is in the database and has a
        # recognized field name.
        return copy.deepcopy(record.fields[field_name])

    # Case 3: The linked record is in the database but does not
    # have a recognized field name.
    dbd_record_type = record_types.get(record.record_type, None)
    if dbd_record_type is None:
        # Record type not in the database?
        return RecordField(
            dtype="invalid",
            name=field_name,
            value="(invalid-record-type)",
            context=_unset_ctx,
        )

    if field_name not in dbd_record_type.fields:
        # Field name invalid
        return RecordField(
            dtype="invalid",
            name=field_name,
            value="(invalid-field)",
            context=_unset_ctx,
        )

    # Record and field found in provided record_types
    dbd_record_field = dbd_record_type.fields[field_name]
    return RecordField(
        dtype=dbd_record_field.type,
        name=field_name,
        value="",
        context=dbd_record_field.context,
    )


def build_database_relations(
    database: Dict[str, RecordInstance],
    record_types: Optional[Dict[str, RecordType]] = None,
    aliases: Optional[Dict[str, str]] = None,
    version: int = 3,
) -> PVRelations:
    """
    Build a dictionary of PV relationships.

    This should not be called often for large databases, as it makes no attempt
    to be computationally efficient.  For repeated usage, cache the result
    of this function and reuse it in future calls to ``graph_links`` and such.

    Parameters
    ----------
    database : dict
        Dictionary of record name to record instance.

    record_types : dict, optional
        The database definitions to use for fields that are not defined in the
        database file.  Dictionary of record type name to RecordType.
        If not specified, the whatrecord-vendored database definition files
        will be used.

    version : int, optional
        Use the old V3 style or new V3 style database grammar by specifying
        3 or 4, respectively.  Defaults to 3.

    Returns
    -------
    info : dict
        Such that: ``info[pv1][pv2] = (field1, field2, info)``
        And in reverse: ``info[pv2][pv1] = (field2, field1, info)``
    """
    if not record_types:
        dbd = Database.from_vendored_dbd(version=version)
        record_types = dbd.record_types

    aliases = aliases or {}
    warned = set()
    by_record = collections.defaultdict(lambda: collections.defaultdict(list))

    for rec1 in database.values():
        rec1_rtype = record_types.get(rec1.record_type, None)
        for field1, link, info in _get_links_for_record(
            rec1, record_types=record_types
        ):
            field1 = copy.deepcopy(field1)
            # field1.context = rec1.context[:1] + field1.context

            if not rec1.has_dbd_info and rec1_rtype:
                field1.update_from_record_type(rec1_rtype)

            if "." in link:
                link, field2_name = link.split(".", 1)
            elif field1.name == "FLNK":
                field2_name = "PROC"
            else:
                field2_name = "VAL"

            rec2_name = aliases.get(link, link)
            rec2 = database.get(rec2_name, None)

            field2 = _field_from_record_relation(
                record=rec2,
                field_name=field2_name,
                link_text=link,
                record_types=record_types,
            )

            if field2 is None:
                continue

            if rec2 is None:
                warned.add(rec2_name)
                logger.debug(
                    "Linked record from %s.%s not in database: %s",
                    rec1.name, field1.name, rec2_name
                )
            else:
                # We may have updated information about the record field;
                # but it's possible this is entirely unnecessary (TODO)
                rec2_type = record_types.get(rec2.record_type, None)
                if rec2_type is not None:
                    field2.update_from_record_type(rec2_type)

            by_record[rec1.name][rec2_name].append((field1, field2, info))
            by_record[rec2_name][rec1.name].append((field2, field1, info))

    return dict(
        (key, dict(inner_dict))
        for key, inner_dict in by_record.items()
    )


def combine_relations(
    dest_relations: PVRelations,
    dest_db: Dict[str, RecordInstance],
    source_relations: PVRelations,
    source_db: Dict[str, RecordInstance],
    record_types: Optional[Dict[str, RecordType]] = None,
    aliases: Optional[Dict[str, str]] = None,
):
    """Combine multiple script relations into one."""
    aliases = aliases or {}

    def get_relation_by_field() -> Tuple[
        str, str, Dict[Tuple[str, str], Tuple[str, str, List[str]]]
    ]:
        for rec1_name, rec2_names in source_relations.items():
            dest_rec1_dict = dest_relations.setdefault(rec1_name, {})
            for rec2_name in rec2_names:
                dest_rec2 = dest_rec1_dict.setdefault(rec2_name, [])
                relation_by_field = {
                    (field1.name, field2.name): (field1, field2, link)
                    for field1, field2, link in dest_rec2
                }
                yield rec1_name, rec2_name, relation_by_field

    # Part 1:
    # Rebuild with new aliases, if available
    # Either set of relations could have referred to aliased names, actual
    # names, or even *both*.

    def alias_to_actual(d):
        # This is kinda expensive, imperfect, and confusing; consider reworking
        for alias_from, alias_to in aliases.items():
            # A -> B
            inner_dict = d.pop(alias_from, None)
            if not inner_dict:
                continue

            # Fix up B <- A first, since it's symmetric
            for inner_name, inner_items in inner_dict.items():
                # d[inner_name][alias_to] += d[inner_name][alias_from]
                d[inner_name].setdefault(alias_to, []).extend(
                    d[inner_name].pop(alias_from)
                )

            if alias_to not in d:
                d[alias_to] = inner_dict
            else:
                # The actual record name is already in the relation dict
                for inner_name, inner_items in inner_dict.items():
                    # d[alias_to][inner_name] += inner_items
                    d[alias_to].setdefault(inner_name, []).extend(inner_items)

    alias_to_actual(dest_relations)
    alias_to_actual(source_relations)

    # Part 1:
    # Merge in new or updated relations from the second set
    for rec1_name, rec2_name, relation_by_field in get_relation_by_field():
        for field1, field2, link in source_relations[rec1_name][rec2_name]:
            key = (field1.name, field2.name)
            existing_link = relation_by_field.get(key, None)
            if not existing_link:
                relation_by_field[key] = (field1, field2, link)
            else:
                existing_field1, existing_field2, _ = existing_link
                existing_field1.update_unknowns(field1)
                existing_field2.update_unknowns(field2)

        dest_relations[rec1_name][rec2_name] = list(relation_by_field.values())

    def get_record(name) -> RecordInstance:
        """Get record from either database."""
        name = aliases.get(name, name)
        try:
            return dest_db.get(name, None) or source_db[name]
        except KeyError:
            raise

    def get_field_info(record, field):
        """Get record definition if available."""
        if field in record.fields:
            return record.fields[field]
        if record_types:
            field_def = record_types[field]
            return RecordField(
                dtype=field_def.type,
                name=field,
                value="",
                context=field_def.context,
            )

        raise KeyError("Field not in database or database definition")

    # Part 2:
    # Update any existing relations in the destination relations with
    # information from the source database
    for rec1_name, rec1 in source_db.items():
        if rec1_name in dest_relations:
            for rec2_name, rec2_items in dest_relations[rec1_name].items():
                # We know rec1 is in the source database, but we don't know
                # where rec2 might be, so use `get_record`.
                try:
                    rec2 = get_record(rec2_name)
                except KeyError:
                    # It's not in this IOC...
                    continue

                def get_items_to_update():
                    for field1, field2, _ in rec2_items:
                        yield (rec1, field1)
                        yield (rec2, field2)
                    for field1, field2, _ in dest_relations[rec2_name][rec1_name]:
                        yield (rec2, field1)
                        yield (rec1, field2)

                for rec, field in get_items_to_update():
                    try:
                        field_info = get_field_info(rec, field.name)
                    except KeyError:
                        logger.debug("Missing field? %s.%s", rec.name, field.name)
                    else:
                        field.update_unknowns(field_info)


def find_record_links(
    database: Dict[str, RecordInstance],
    starting_records: List[str],
    relations: Optional[PVRelations] = None,
    record_types: Optional[Dict[str, RecordType]] = None,
) -> Generator[LinkInfo, None, None]:
    """
    Get all related record links from a set of starting records.

    All starting records will be included, along with any other records that
    are linked to from there.

    Parameters
    ----------
    database : dict
        Dictionary of record name to record instance.

    starting_records : list of str
        Record names

    relations : dict, optional
        Pre-built PV relationship dictionary.  Generated from database
        if not provided.

    record_types : dict, optional
        The database definitions to use for fields that are not defined in the
        database file.  Dictionary of record type name to RecordType.

    Yields
    ------
    link_info : LinkInfo
        Link info
    """
    checked = []

    if relations is None:
        relations = build_database_relations(database, record_types=record_types)

    records_to_check = list(starting_records)

    while records_to_check:
        rec1 = database.get(records_to_check.pop(), None)
        if rec1 is None:
            continue

        checked.append(rec1.name)
        logger.debug("--- record %s ---", rec1.name)

        for rec2_name, fields in relations.get(rec1.name, {}).items():
            if rec2_name in checked:
                continue

            rec2 = database.get(rec2_name, None)
            if rec2 is None:
                continue

            for field1, field2, info in fields:
                if rec2_name not in checked and rec2_name not in records_to_check:
                    records_to_check.append(rec2_name)

                li = LinkInfo(
                    record1=rec1,
                    field1=field1,
                    record2=rec2,
                    field2=field2,
                    info=info,
                )

                logger.debug("Link %s", li)
                yield li


class RecordLinkGraph(_GraphHelper):
    """Record link graph."""
    # TODO: create node and color when not in database?

    database: Optional[Database]
    starting_records: List[str]
    newline: str = '\n'
    header_format: str = '{name} ({rtype})'
    field_format: str = '{field}: "{value}"'
    text_format: str = (
        """
        <TABLE BORDER="0" CELLBORDER="0">
        <TR>
            <TD>{rtype}</TD>
            <TD><b>{name}</b></TD>
        </TR>
        {field_lines}
        </TABLE>
        """.strip()
    )
    sort_fields: bool
    show_empty: bool
    relations: Optional[PVRelations]
    record_types: Dict[str, RecordType]
    default_edge_kwargs: Dict[str, Any] = {
        "style": "solid",
    }

    edge_kwargs: ClassVar[Dict[str, Dict[str, str]]] = {
        "style": {
            "PP": "bold",
            "CPP": "bold",
            "CP": "bold",
        },
        # "color": {
        #     "MS": "",
        #     "MSS": "",
        #     "MSI": "",
        # }
    }

    # Edge colors to cycle through: black is reserved for FLNK.
    edge_colors: ClassVar[List[str]] = textwrap.dedent(
        """\
        brown
        darkblue
        darkcyan
        darkgreen
        darkmagenta
        darkred
        olive
        """.rstrip()
    ).split()

    def __init__(
        self,
        database: Optional[Union[Database, Dict[str, RecordInstance]]] = None,
        starting_records: Optional[List[str]] = None,
        header_format: Optional[str] = None,
        field_format: Optional[str] = None,
        text_format: Optional[str] = None,
        sort_fields: bool = True,
        show_empty: bool = True,
        relations: Optional[PVRelations] = None,
        record_types: Optional[Dict[str, RecordType]] = None,
    ):
        super().__init__()
        self._built = False
        self.database = None
        self.starting_records = starting_records or []
        self.header_format = header_format or type(self).header_format
        self.field_format = field_format or type(self).field_format
        self.text_format = text_format or type(self).text_format
        self.sort_fields = sort_fields
        self.show_empty = show_empty
        self.relations = relations

        if database is not None:
            self.add_database(database)

        if record_types is not None:
            if self.database is None:
                self.database = Database()
            self.database.record_types.update(record_types)

        self.relations = relations

    def get_node(
        self, label: str, text: Optional[str] = None
    ) -> GraphNode:
        node = super().get_node(label, text=text)
        if not node.metadata:
            node.metadata["fields"] = {}
        return node

    def add_database(self, database: Union[Dict[str, RecordInstance], Database]):
        """Add records from the given database to the graph."""
        if self.database is None:
            if isinstance(database, Database):
                self.database = database.shallow_copy()
            else:
                self.database = Database(
                    records=database,
                )
        elif isinstance(database, Database):
            self.database.append(database)
        else:
            for record in database.values():
                self.database.add_or_update_record(record)

        self._built = False
        self.relations = None

    def build(self):
        """Build the graph from the provided databases."""
        if self.database is None:
            raise ValueError("No database or records were added")

        edge_cache = set()

        def add_edge(
            source: GraphNode,
            destination: GraphNode,
            source_port: Optional[str] = None,
            dest_port: Optional[str] = None,
            **options
        ) -> None:
            key = (source.id, destination.id)
            if key in edge_cache:
                return
            edge_cache.add(key)
            # Avoid bottleneck of checking the edge list... ?
            self.edges.append(
                GraphEdge(
                    source,
                    destination,
                    source_port=source_port,
                    destination_port=dest_port,
                    options=options
                )
            )

        edge_color_idx = 0
        for li in find_record_links(
            self.database.records, self.starting_records, relations=self.relations
        ):
            src = self.get_node(li.record1.name, text=" ")
            dest = self.get_node(li.record2.name, text=" ")

            for field, node in [(li.field1, src), (li.field2, dest)]:
                if field.name == "PROC":
                    ...
                elif field.value or self.show_empty:
                    node.metadata["fields"][field.name] = (
                        f'<TR>'
                        f'<TD PORT="{field.name}_name" BORDER="1">'
                        f'<B>{field.name}</B>'
                        f'</TD>'
                        f'<TD PORT="{field.name}_value" BORDER="1">'
                        f'{field.value}'
                        f'</TD>'
                        f'</TR>'
                    )

            if li.field1.dtype == "DBF_INLINK" or li.field2.dtype == "DBF_OUTLINK":
                src, dest = dest, src
                li.field1, li.field2 = li.field2, li.field1

            logger.debug("New edge %s -> %s", src, dest)

            edge_kw = dict(self.default_edge_kwargs)
            for key, to_find in self.edge_kwargs.items():
                for match, value in to_find.items():
                    if match in li.info:
                        edge_kw[key] = value
                        break

            if (src.id, dest.id) in edge_cache:
                continue

            if li.field1.dtype == "DBF_FWDLINK":
                # edge_kw["taillabel"] = "FLNK"
                add_edge(
                    src,
                    dest,
                    color="black",
                    source_port=li.field1.name,
                    **edge_kw
                )
            elif li.field2.dtype == "DBF_FWDLINK":
                # edge_kw["taillabel"] = "FLNK"
                add_edge(
                    dest,
                    src,
                    color="black",
                    source_port=li.field2.name,
                    **edge_kw
                )
            else:
                # edge_kw["taillabel"] = f"{li.field1.name}"
                # edge_kw["headlabel"] = f"{li.field2.name}"
                if li.info:
                    edge_kw["xlabel"] = f"\n{' '.join(li.info)}"

                edge_color_idx += 1
                color = self.edge_colors[edge_color_idx % len(self.edge_colors)]

                add_edge(
                    src,
                    dest,
                    source_port=li.field1.name,
                    dest_port=li.field2.name,
                    color=color,
                    **edge_kw,
                )

        if not self.nodes:
            # No relationship found; at least show the records
            for rec_name in self.starting_records:
                if rec_name in self.database.records:
                    self.get_node(rec_name)

        for node in self.nodes.values():
            # if self.sort_fields:
            #     node.text = "\n".join(sorted(node.text.splitlines()))
            rec = self.database.records[node.label]
            if rec.aliases:
                sub_header = html.escape(f"\nAlias: {', '.join(rec.aliases)}")
                sub_header = f"""<TR><TD>{sub_header}</TD></TR>"""
            else:
                sub_header = ""

            fields = list(sorted(node.metadata["fields"].items()))
            field_text = "\n".join(text for _, text in fields)
            node.text = self.text_format.format(
                rtype=html.escape(rec.record_type),
                name=html.escape(rec.name),
                field_lines=(sub_header + field_text).strip(),
            )

        self._built = True

    @property
    def graphed_records(self) -> List[str]:
        """All graphed record names (i.e., node labels)."""
        return [
            node.label
            for node in self.nodes.values()
        ]

    def _ready_for_digraph(self, graph: gv.Digraph):
        """Hook when the user calls ``to_digraph``."""
        if not self._built:
            self.build()

        all_match = set(self.graphed_records) <= set(self.starting_records)
        for node in self.nodes.values():
            node.highlighted = (
                not all_match and
                node.label in self.starting_records
            )


def graph_links(
    database: Union[Database, Dict[str, RecordInstance]],
    starting_records: List[str],
    header_format: Optional[str] = None,
    field_format: Optional[str] = None,
    text_format: Optional[str] = None,
    sort_fields: bool = True,
    show_empty: bool = True,
    relations: Optional[PVRelations] = None,
    record_types: Optional[Dict[str, RecordType]] = None,
) -> RecordLinkGraph:
    """
    Create a graphviz digraph of record links.

    All starting records will be included, along with any other records that
    are linked to from there - if available in the database.

    Parameters
    ----------
    database : dict
        Dictionary of record name to record instance.
    starting_records : list of str
        Record names
    graph : graphviz.Graph, optional
        Graph instance to use. New one created if not specified.
    engine : str, optional
        Graphviz engine (dot, fdp, etc)
    field_format : str, optional
        Format string for fields (keys: field, value, attr)
    sort_fields : bool, optional
        Sort list of fields
    text_format : str, optional
        Text format for full node (keys: header, field_lines)
    show_empty : bool, optional
        Show empty fields
    font_name : str, optional
        Font name to use for all nodes and edges
    relations : dict, optional
        Pre-built PV relationship dictionary.  Generated from database
        if not provided.
    record_types : dict, optional
        The database definitions to use for fields that are not defined in the
        database file.  Dictionary of record type name to RecordType.  Only
        used for determining script relations if not specified.

    Returns
    -------
    graph : RecordLinkGraph
    """

    return RecordLinkGraph(
        database=database,
        starting_records=starting_records,
        header_format=header_format,
        field_format=field_format,
        text_format=text_format,
        sort_fields=sort_fields,
        show_empty=show_empty,
        relations=relations,
        record_types=record_types,
    )


def build_script_relations(
    database: Dict[str, RecordInstance],
    by_record: PVRelations,
    limit_to_records: Optional[List[str]] = None
) -> ScriptPVRelations:
    """
    Build a relational map of IOCs that include inter-IOC PV relationships.

    Parameters
    ----------
    database : Dict[str, RecordInstance]
        The full composite database.
    by_record : PVRelations
        PV relationships, as generated by ``build_database_relations``.
        This is a mapping of PV1 to {"PV2": [FieldRelation, ...]}.
    limit_to_records : List[str], optional
        Limit the relationships to only the provided records.  If not set,
        defaults to all records in the composite database.

    Returns
    -------
    ScriptPVRelations
    """
    if not limit_to_records:
        record_items = by_record.items()
    else:
        record_items = [
            (name, by_record[name]) for name in limit_to_records
            if name in by_record
        ]

    def get_owner(rec):
        if not rec:
            return "unknown"

        if rec.owner and rec.owner != "unknown":
            return rec.owner

        if rec.context:
            return rec.context[0].name

        return "unknown"

    by_script = collections.defaultdict(lambda: collections.defaultdict(set))
    for rec1_name, list_of_rec2s in record_items:
        rec1 = database.get(rec1_name, None)
        for rec2_name in list_of_rec2s:
            rec2 = database.get(rec2_name, None)

            owner1 = get_owner(rec1)
            owner2 = get_owner(rec2)
            if owner1 != owner2:
                by_script[owner2][owner1].add(rec2_name)
                by_script[owner1][owner2].add(rec1_name)

    return {
        owner1: {
            owner2: list(items)
            for owner2, items in owner2s.items()
        }
        for owner1, owner2s in by_script.items()
    }


class ScriptLinkGraph(_GraphHelper):
    """
    Script link graph (i.e., inter-IOC record links).

    Parameters
    ----------
    database : dict
        Dictionary of record name to record instance.
    starting_records : list of str
        Record names
    sort_fields : bool, optional
        Sort list of fields
    show_empty : bool, optional
        Show empty fields
    relations : dict, optional
        Pre-built PV relationship dictionary.  Generated from database
        if not provided.
    script_relations : dict, optional
        Pre-built script relationship dictionary.  Generated from database if
        not provided.
    record_types : dict, optional
        The database definitions to use for fields that are not defined in the
        database file.  Dictionary of record type name to RecordType.  Only
        used for determining script relations if not specified.
    """
    # TODO: create node and color when not in database?

    newline: str = '<br align="center"/>'
    _built: bool
    limit_to_records: List[str]
    script_relations: Optional[ScriptPVRelations]

    def __init__(
        self,
        database: Optional[Union[Database, Dict[str, RecordInstance]]] = None,
        limit_to_records: Optional[List[str]] = None,
        relations: Optional[PVRelations] = None,
        script_relations: Optional[ScriptPVRelations] = None,
        record_types: Optional[Dict[str, RecordType]] = None,
    ):
        super().__init__()
        self.database = None
        self.limit_to_records = limit_to_records or []
        self._built = False

        if database is not None:
            self.add_database(database)

        if record_types is not None:
            if self.database is None:
                self.database = Database()
            self.database.record_types.update(record_types)

        self.relations = relations
        self.script_relations = script_relations

    def add_database(self, database: Union[Dict[str, RecordInstance], Database]):
        """Add records from the given database to the graph."""
        if self.database is None:
            if isinstance(database, Database):
                self.database = database.shallow_copy()
            else:
                self.database = Database(
                    records=database,
                )
        elif isinstance(database, Database):
            self.database.append(database)
        else:
            for record in database.values():
                self.database.add_or_update_record(record)

        self._built = False
        self.relations = None
        self.script_relations = None

    def build(self):
        """Build the graph from the provided databases."""
        if self.database is None:
            raise ValueError("No database or records were added")

        if self.relations is None:
            self.relations = build_database_relations(
                self.database.records,
                record_types=self.database.record_types,
            )

        if self.script_relations is None:
            self.script_relations = build_script_relations(
                database=self.database.records,
                by_record=self.relations,
                limit_to_records=self.limit_to_records,
            )

        for script_a, script_a_relations in self.script_relations.items():
            self.get_node(script_a, text=script_a)
            for script_b in script_a_relations:
                if script_b in self.nodes:
                    continue
                self.get_node(script_b, text=script_b)

                inter_lines = (
                    [f"<b>{script_a}</b>", ""]
                    + list(sorted(self.script_relations[script_a][script_b]))
                    + [""]
                    + [f"<b>{script_b}</b>", ""]
                    + list(sorted(self.script_relations[script_b][script_a]))
                )
                inter_node = f"{script_a}<->{script_b}"
                self.get_node(inter_node, text="\n".join(inter_lines))
                self.add_edge(script_a, inter_node)
                self.add_edge(inter_node, script_b)

        if not self.nodes:
            # No relationship found; at least show the records
            for rec_name in self.limit_to_records or []:
                if rec_name in self.database.records:
                    self.get_node(rec_name)

        self._built = True

    def _ready_for_digraph(self, graph: gv.Digraph):
        """Hook when the user calls ``to_digraph``."""
        if not self._built:
            self.build()
        for node in self.nodes.values():
            node.highlighted = node.label in self.limit_to_records


def graph_script_relations(
    database: Dict[str, RecordInstance],
    limit_to_records: Optional[List[str]] = None,
    relations: Optional[PVRelations] = None,
    script_relations: Optional[ScriptPVRelations] = None,
    record_types: Optional[Dict[str, RecordType]] = None,
) -> ScriptLinkGraph:
    """
    Create a graphviz digraph of script links (i.e., inter-IOC record links).

    Parameters
    ----------
    database : dict
        Dictionary of record name to record instance.
    starting_records : list of str
        Record names
    relations : dict, optional
        Pre-built PV relationship dictionary.  Generated from database
        if not provided.
    script_relations : dict, optional
        Pre-built script relationship dictionary.  Generated from database if
        not provided.
    record_types : dict, optional
        The database definitions to use for fields that are not defined in the
        database file.  Dictionary of record type name to RecordType.  Only
        used for determining script relations if not specified.

    Returns
    -------
    graph : ScriptLinkGraph
    """
    return ScriptLinkGraph(
        database=database,
        limit_to_records=limit_to_records,
        relations=relations,
        script_relations=script_relations,
        record_types=record_types,
    )



"""
This module is not intended to be imported and used in Python.  It's a helper
script to be paired with the GNU debugger (gdb).

Usage:
    $ gdb --batch-silent --command gdb_binary_info.py --args /path/to/bin/linux-x86_64/ioc_binary
"""
from __future__ import print_function

if __name__ != "__main__":
    raise RuntimeError(__doc__)

import json  # noqa
import os  # noqa
import sys  # noqa
import tempfile  # noqa

import gdb  # noqa

try:
    gdb.execute("set startup-with-shell off")
except gdb.error:
    # Comes in a later version
    pass


def get_all_symbol_names():
    try:
        msymbols = gdb.execute("maint print msymbols", to_string=True)
    except gdb.error:
        with tempfile.NamedTemporaryFile(mode="wt") as temp:
            gdb.execute("maint print msymbols " + temp.name, to_string=True)
            with open(temp.name, "rt") as f:
                msymbols = f.read()

    for line in msymbols.splitlines():
        line = line.strip()
        if not line.count(" ") > 3:
            continue

        symbol_name = line.split()[3]
        yield line, symbol_name


def parse_version_string(ver):
    return ver.split(" ")[2].rstrip(",")


def get_version_info():
    try:
        version = gdb.parse_and_eval("pVersionCAC")
        return parse_version_string(str(version))
    except gdb.error:
        return None


OLD_GDB = not hasattr(gdb, "lookup_static_symbol")

if not OLD_GDB:
    # TODO: unify or just remove old gdb support?
    def get_variables():
        defs = gdb.lookup_static_symbol("vardefs")
        if not defs:
            return {}

        variables = {}
        idx = 0
        while True:
            var = defs.value()[idx]
            name = var["name"]
            if int(name.dereference().address) == 0:
                break
            variables[name.string()] = {
                "name": name.string(),
                "type": str(var["type"]),
                "value": None,
            }
            idx += 1
        return variables

    def find_funcdefs(symbols=None):
        variables = gdb.execute("info variables -q -t iocshFuncDef", to_string=True)
        funcdef_names = [
            line.split()[-1].rstrip(";")
            for line in variables.splitlines()
            if 'iocshFuncDef' in line
        ]

        for funcdef_name in funcdef_names:
            sym = gdb.lookup_static_symbol(funcdef_name)
            try:
                val = sym.value()
            except Exception:
                continue

            try:
                name = val["name"].string()
            except gdb.error:
                # optimized out; skip this one entirely
                continue

            yield name, sym, val

else:
    def get_variables():
        # TODO
        return {}

    def find_funcdefs(symbols=None):
        for line, symbol_name in (symbols or ALL_SYMBOLS):
            sym, is_method_field = gdb.lookup_symbol(symbol_name)
            if is_method_field:
                continue

            if sym is None or str(sym.type) not in ("const iocshFuncDef", "iocshFuncDef"):
                continue

            try:
                val = sym.value()
            except Exception:
                continue

            try:
                val["name"].string()
            except gdb.error:
                # optimized out; skip this one entirely
                continue

            yield symbol_name, sym, val


def get_symbol_context(sym):
    """Get the whatrecord-format context for the given gdb symbol."""
    return (os.path.abspath(sym.symtab.fullname()), sym.line)


def get_commands():
    def by_name(item):
        name, _, _, = item
        return name

    for name, funcdef_sym, funcdef in sorted(find_funcdefs(), key=by_name):
        try:
            context = get_symbol_context(funcdef_sym)
        except gdb.error:
            context = None

        command = {
            "name": name,
            "usage": None,
            "context": (context, ) if context else None,
            "args": [],
        }
        try:
            usage = funcdef["usage"]
            if isinstance(usage, gdb.Value):
                usage = usage.string()
        except Exception:
            pass
        else:
            command["usage"] = usage

        for idx in range(int(funcdef["nargs"])):
            arg = funcdef["arg"][idx].dereference()
            command["args"].append({
                "name": arg["name"].string(),
                "type": str(arg["type"]),
            })
        yield name, command


if OLD_GDB:
    # gdb.execute("set logging off")
    # Do this before "start" - static symbols only for now
    ALL_SYMBOLS = tuple(sorted(get_all_symbol_names()))
    gdb.execute("start")

try:
    info = {
        "commands": dict(get_commands()),
        "base_version": get_version_info(),
        "variables": get_variables(),
        "error": None,
    }
except Exception as ex:
    info = {
        "commands": {},
        "base_version": None,
        "variables": {},
        "error": "{0}: {1}".format(type(ex).__name__, ex),
    }
finally:
    if OLD_GDB:
        gdb.execute("kill")


# Note: --batch-silent will eat up all output of this script, too; so
# use the original sys.__stdout__
print(json.dumps(info, indent=4), file=sys.__stdout__)

# gdb.execute("quit")



"""
happi whatrecord plugin

Match your happi devices to IOCs and records.
"""
from __future__ import annotations

import argparse
import collections
import fnmatch
import json
import logging
import typing
from dataclasses import dataclass
from typing import Dict, Generator, List, Tuple, Type, TypeVar, Union

import apischema

from ..server.common import PluginResults
from ..util import get_file_sha256
from .util import suppress_output_decorator

try:
    import happi
    import happi.backends
    import happi.backends.json_db
    import ophyd
    from ophyd.ophydobj import OphydObject
    from ophyd.signal import EpicsSignalBase
except ImportError as ex:
    raise ImportError(f"Dependency required for happi plugin unavailable: {ex}")


logger = logging.getLogger(__name__)

# Stash the description for later usage by the CLI interface.
DESCRIPTION = __doc__.strip()
CriteriaDict = Dict[str, Union[float, str]]
T = TypeVar("T")


HappiItem = Dict[str, Union[str, List[str], Dict[str, str]]]  # kinda...

# TODO: this can be improved
SIGNAL_CLASSES = (
    EpicsSignalBase,
    ophyd.EpicsMotor,
    ophyd.EpicsScaler,
)


@dataclass
class HappiPluginResults(PluginResults):
    # Could potentially further specify metadata_by_key or metadata
    ...


@dataclass
class HappiRecordInfo:
    name: str
    kind: str
    signal: str


def get_all_devices(
    client: happi.Client = None,
) -> Generator[ophyd.Device, None, None]:
    """
    Get all devices from a given happi client.

    Parameters
    ----------
    client : happi.Client, optional
        The happi client to use.  Defaults to using one from the environment
        configuration.

    Yields
    ------
    ophyd.Device
    """
    if client is None:
        client = happi.Client.from_config()

    if hasattr(client.backend, "_load_or_initialize"):
        # HACK: TODO: avoid re-re-re-reading the JSON database; happi needs
        # some work.
        loaded_database = client.backend._load_or_initialize()
        client.backend._load_or_initialize = lambda: loaded_database

    for dev in client:
        try:
            obj = client[dev].get()
        except Exception:
            logger.exception("Failed to instantiate device: %s", dev)
        else:
            yield obj


def get_devices_by_criteria(
    search_criteria: CriteriaDict,
    *,
    client: happi.Client = None,
    regex: bool = True,
) -> Generator[ophyd.Device, None, None]:
    """
    Get all devices from a given happi client.

    Parameters
    ----------
    search_criteria : dict
        Dictionary of ``{'happi_key': 'search_value'}``.

    client : happi.Client, optional
        The happi client to use.  Defaults to using one from the environment
        configuration.

    Yields
    ------
    ophyd.Device
    """
    if client is None:
        client = happi.Client.from_config()

    search_method = client.search_regex if regex else client.search
    for item in search_method(**search_criteria):
        try:
            obj = item.get()
        except Exception:
            logger.exception("Failed to instantiate device: %s", obj)
        else:
            yield obj


def get_components_matching(
    obj: ophyd.Device,
    predicate: callable,
) -> Generator[OphydObject, None, None]:
    """
    Find signals of a specific type from a given ophyd Device.

    Parameters
    ----------
    obj : ophyd.Device
        The ophyd Device.

    predicate : callable
        Callable that should return True on a match.

    Yields
    ------
    obj : OphydObject
    """
    for name, dev in obj.walk_subdevices(include_lazy=True):
        try:
            included = predicate(dev)
        except Exception:
            logger.exception("Failed to check predicate against %s %s", name, dev)
        else:
            if included:
                yield dev

    for walk in obj.walk_signals(include_lazy=True):
        try:
            included = predicate(walk.item)
        except Exception:
            logger.exception("Failed to check predicate against %s", walk)
        else:
            if included:
                yield walk.item


def patch_and_use_dummy_shim():
    """
    Hack ophyd and its dummy shim.  We don't want _any_ control-layer
    connections being made while we're looking for signals.

    Warning
    -------
    Under no circumstances should this be used in a production environment
    where you intend to actually _use_ ophyd for its intended purpose.
    """
    ophyd.Device.lazy_wait_for_connection = False

    def _no_op(*args, **kwargs):
        ...

    class _PVStandIn:
        _reference_count = 0

        def __init__(self, pvname, *args, **kwargs):
            self.pvname = pvname
            self.connected = True

        add_callback = _no_op
        remove_callback = _no_op
        clear_callbacks = _no_op
        get = _no_op
        put = _no_op
        get_with_metadata = _no_op
        wait_for_connection = _no_op

    def get_pv(pvname, *args, **kwargs):
        return _PVStandIn(pvname)

    from ophyd import _dummy_shim

    _dummy_shim.get_pv = get_pv
    _dummy_shim.release_pvs = _no_op
    ophyd.set_cl("dummy")


def find_signals(
    criteria: CriteriaDict,
    signal_class: Type[T],
) -> Generator[Tuple[OphydObject, T], None, None]:
    """
    Find all signal metadata that match the given criteria.

    Yields
    ------
    root : OphydObject
        The root OphydObject - may be the same as signal.  This is guaranteed
        to have happi metadata on it as the ``md`` attribute.

    sig : signal_class
        Signals of type `signal_class`.
    """

    patch_and_use_dummy_shim()

    def is_of_class(obj):
        return isinstance(obj, signal_class)

    if not criteria:
        devices = get_all_devices()
    else:
        devices = get_devices_by_criteria(criteria)

    for dev in devices:
        if not hasattr(dev, "md"):
            # This should be a guarantee, but check just in case.
            continue

        for sig in get_components_matching(dev, predicate=is_of_class):
            yield dev, sig
        # Top-level devices are OK too
        if isinstance(dev, signal_class):
            yield dev, dev


def find_signal_metadata_pairs(
    criteria: CriteriaDict,
) -> Generator[Tuple[OphydObject, str, EpicsSignalBase], None, None]:
    """
    Find all signal metadata that match the given criteria.

    Yields
    ------
    root : OphydObject
        The root OphydObject - may be the same as signal.  This is guaranteed
        to have happi metadata on it as the ``md`` attribute.

    pvname : str
        The PV name.

    signal : EpicsSignal, EpicsSignalRO, ...
        The signal instance, one of ``SIGNAL_CLASSES``.
    """

    attributes = [
        "pvname",
        "setpoint_pvname",
    ]

    for root, sig in find_signals(criteria, signal_class=SIGNAL_CLASSES):
        found_pvs = set()
        for attr in attributes:
            pvname = getattr(sig, attr, None)
            if pvname is not None and isinstance(pvname, str):
                found_pvs.add(pvname)

        if found_pvs:
            for pvname in sorted(found_pvs):
                yield root, pvname, sig
        else:
            # Then it's the prefix - possibly
            pvname = getattr(sig, "prefix", None)
            if pvname is not None:
                yield root, pvname, sig


def _parse_criteria(criteria_string: str) -> CriteriaDict:
    """
    Parse search criteria into a dictionary of ``{key: value}``.

    Converts floating point values to float.
    """
    search_args = {}
    for user_arg in criteria_string:
        if "=" in user_arg:
            criteria, value = user_arg.split("=", 1)
        else:
            criteria = "name"
            value = user_arg

        if criteria in search_args:
            logger.warning(
                "Received duplicate search criteria %s=%r (was %r)",
                criteria,
                value,
                search_args[criteria],
            )
            continue

        try:
            value = float(value)
        except ValueError:
            value = fnmatch.translate(value)

        search_args[criteria] = value

    return search_args


def _get_argparser(parser: typing.Optional[argparse.ArgumentParser] = None):
    if parser is None:
        parser = argparse.ArgumentParser(description=DESCRIPTION)

    parser.add_argument(
        "search_criteria", nargs="*", help="Search criteria: field=value"
    )
    parser.add_argument(
        "-p", "--pretty", action="store_true", help="Pretty JSON output"
    )
    return parser


def get_all_happi_items(client: happi.Client) -> Generator[happi.SearchResult, None, None]:
    """
    Get all happi search results from the given client.

    Parameters
    ----------
    client : happi.Client
        The happi client instance.

    Yields
    ------
    SearchResult
        The happi item wrapped in a SearchResult.
    """
    for name in client:
        try:
            yield client[name]
        except Exception:
            logger.debug("Failed to get happi item: %s", name, exc_info=True)


@suppress_output_decorator
def main(search_criteria: str, pretty: bool = False):
    client = happi.Client.from_config()

    files_to_monitor = {}

    if isinstance(client.backend, happi.backends.json_db.JSONBackend):
        files_to_monitor[client.backend.path] = get_file_sha256(
            client.backend.path
        )

    results = HappiPluginResults(
        files_to_monitor=files_to_monitor,
        record_to_metadata_keys=collections.defaultdict(list),
        metadata_by_key={
            result.item["name"]: dict(result.item)
            for result in get_all_happi_items(client)
        },
        execution_info={},
    )

    criteria = _parse_criteria(search_criteria)
    for root, record, sig in find_signal_metadata_pairs(criteria):
        if "." in record:
            record, *_ = record.split(".")

        happi_md = root.md

        if happi_md.name not in results.record_to_metadata_keys[record]:
            results.record_to_metadata_keys[record].append(happi_md.name)
            try:
                md = results.metadata_by_key[happi_md.name]
            except KeyError:
                # Mismatched _id and name
                continue

            if "_whatrecord" not in md:
                md["_whatrecord"] = {"records": []}
            if sig.root is not root:
                # TODO: this would not work for nested GroupDevice components.
                dotted_name = ".".join((sig.root.attr_name, sig.dotted_name))
            else:
                dotted_name = sig.dotted_name
            md["_whatrecord"]["records"].append(
                HappiRecordInfo(
                    name=record,
                    kind=str(sig.kind),
                    signal=dotted_name,
                )
            )

    return results


def _cli_main():
    parser = _get_argparser()
    args = parser.parse_args()
    results = main(**vars(args))
    json_results = apischema.serialize(results)
    dump_args = {"indent": 4} if args.pretty else {}
    print(json.dumps(json_results, sort_keys=True, **dump_args))


if __name__ == "__main__":
    _cli_main()



"""
`whatrecord` is the top-level command for accessing various subcommands.

Try::

"""

import argparse
import asyncio
import importlib
import inspect
import logging

import whatrecord

DESCRIPTION = __doc__
RETURN_VALUE = None

# The whatrecord.bin modules are listed here and imported dynamically such
# that any missing dependencies for that command will not stop other
# entrypoints from working:
MODULES = (
    "deps",
    "graph",
    "info",
    "iocmanager_loader",
    "lint",
    "parse",
    "server",
)


def _try_import(module):
    relative_module = f".{module}"
    return importlib.import_module(relative_module, "whatrecord.bin")


def _build_commands():
    global DESCRIPTION
    result = {}
    unavailable = []

    for module in sorted(MODULES):
        subcommand = module.replace("_", "-")
        try:
            mod = _try_import(module)
        except Exception as ex:
            unavailable.append((module, ex))
        else:
            result[subcommand] = (mod.build_arg_parser, mod.main)
            DESCRIPTION += f"\n    $ whatrecord {subcommand} --help"

    if unavailable:
        DESCRIPTION += "\n\n"

        for module, ex in unavailable:
            DESCRIPTION += (
                f"\nWARNING: whatrecord {module!r} is unavailable due to:"
                f"\n\t{ex.__class__.__name__}: {ex}"
            )

    return result


COMMANDS = _build_commands()


def main():
    top_parser = argparse.ArgumentParser(
        prog="whatrecord",
        description=DESCRIPTION,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    top_parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=whatrecord.__version__,
        help="Show the whatrec version number and exit.",
    )

    top_parser.add_argument(
        "--log",
        "-l",
        dest="log_level",
        default="INFO",
        type=str,
        help="Python logging level (e.g. DEBUG, INFO, WARNING)",
    )

    subparsers = top_parser.add_subparsers(help="Possible subcommands")
    for command_name, (build_func, main) in COMMANDS.items():
        sub = subparsers.add_parser(command_name)
        build_func(sub)
        sub.set_defaults(func=main)

    args = top_parser.parse_args()
    kwargs = vars(args)
    log_level = kwargs.pop("log_level")

    logger = logging.getLogger("whatrecord")
    logger.setLevel(log_level)
    logging.basicConfig()

    if hasattr(args, "func"):
        func = kwargs.pop("func")
        logger.debug("%s(**%r)", func.__name__, kwargs)

        global RETURN_VALUE
        if inspect.iscoroutinefunction(func):
            loop = asyncio.get_event_loop()
            RETURN_VALUE = loop.run_until_complete(func(**kwargs))
        else:
            RETURN_VALUE = func(**kwargs)
            logger.debug(
                "%s Return value available as RETURN_VALUE: %s",
                func.__name__, RETURN_VALUE
            )
    else:
        top_parser.print_help()


if __name__ == "__main__":
    main()



"""
V3 Access Security file parsing.

Documentation from the application developer's guide are interspersed here
and in the classes below.

```
A brief summary of the Functional Requirements is:
    * Each field of each record type is assigned an access security level.
    * Each record instance is assigned to a unique access security group.
    * Each user is assigned to one or more user access groups.
    * Each node is assigned to a host access group.

For each access security group a set of access rules can be defined. Each rule
specifies:
    -  Access security level
    -  READ or READ/WRITE access.
    -  An optional list of User Access Groups or * meaning anyone.
    -  An optional list of Host Access Groups or * meaning anywhere.
    -  Conditions based on values of process variables
```
"""

from __future__ import annotations

import logging
import pathlib
from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, List, Optional, Tuple, Union

import apischema
import lark

from . import transformer
from .common import (FullLoadContext, RecordInstance, ShellStateHandler,
                     StringWithContext, context_from_lark_token)
from .macro import MacroContext
from .util import get_bytes_sha256

MODULE_PATH = pathlib.Path(__file__).parent.resolve()
logger = logging.getLogger(__name__)


@dataclass
class UserAccessGroup:
    """
    User Access Group.

    This is a list of user names. The list may be empty. A user name may appear
    in more than one UAG. To match, a user name must be identical to the user
    name read by the CA client library running on the client machine. For
    vxWorks clients, the user name is usually taken from the user field of the
    boot parameters.
    """
    context: FullLoadContext
    comments: str
    name: str
    users: List[str]


@dataclass
class HostAccessGroup:
    """
    Host Access Group.

    This is a list of host names. It may be empty. The same host name can
    appear in multiple HAGs. To match, a host name must match the host name
    read by the CA client library running on the client machine; both names are
    converted to lower case before comparison however. For vxWorks clients, the
    host name is usually taken from the target name of the boot parameters.
    """
    context: FullLoadContext
    comments: str
    name: str
    hosts: List[str]


@apischema.fields.with_fields_set
@dataclass
class AccessSecurityRule:
    """
    Access Security Configuration rule, which defines access permissions.

    <level> must be 0 or 1. Permission for a level 1 field implies permission
    for level 0 fields.

    The permissions are NONE, READ, and WRITE. WRITE permission implies READ
    permission. The standard EPICS record types have all fields set to level 1
    except for VAL, CMD (command), and RES (reset). An optional argument
    specifies if writes should be trapped. See the section below on trapping
    Channel Access writes for how this is used. If not given the default is
    NOTRAPWRITE.

    UAG specifies a list of user access groups that can have the access
    privilege. If UAG is not defined then all users are allowed.

    HAG specifies a list of host access groups that have the access privilege.
    If HAG is not defined then all hosts are allowed.

    CALC is just like the CALC field of a calculation record except that the
    result must evaluate to TRUE or FALSE. The rule only applies if the
    calculation result is TRUE, where the actual test for TRUE is
    (0.99 < result < 1.01).

    Anything else is regarded as FALSE and will cause the rule to be ignored.
    Assignment statements are not permitted in CALC expressions here.
    """
    context: FullLoadContext
    comments: str
    level: int
    options: str
    log_options: Optional[str] = None
    users: Optional[List[str]] = None
    hosts: Optional[List[str]] = None
    calc: Optional[str] = None


@dataclass
class AccessSecurityGroup:
    """An access security group."""
    context: FullLoadContext
    comments: str
    name: str
    inputs: Dict[str, str] = field(default_factory=dict)
    rules: List[AccessSecurityRule] = field(default_factory=list)


@lark.visitors.v_args(inline=True)
class _AcfTransformer(lark.visitors.Transformer):
    fn: str
    contents_hash: str
    comments: List[str]
    hosts: Dict[str, HostAccessGroup]
    users: Dict[str, UserAccessGroup]
    groups: Dict[str, AccessSecurityGroup]
    inputs: Dict[str, str]
    calc: Tuple[lark.Token, str]

    def __init__(self, fn, contents_hash: str, comments: List[str]):
        super().__init__(visit_tokens=False)
        self.fn = str(fn)
        self.contents_hash = contents_hash
        self.comments = comments
        self.hosts = {}
        self.users = {}
        self.groups = {}
        self.rule_info = {}
        self.rules = []
        self.inputs = {}
        self.calc = None

    def get_header_comments(self) -> str:
        """Get all comments at the top of the file."""
        results = []
        lineno = 0
        for comment in self.comments:
            if comment.context[0].line == lineno + 1:
                results.append(str(comment).lstrip("# "))
                lineno += 1
            else:
                break

        return "\n".join(results)

    def get_recent_comments(self, lineno: int) -> str:
        """Get all comments just before the given line."""
        results = []
        for comment in reversed(self.comments):
            comment_line = comment.context[0].line
            if comment_line == lineno - 1:
                results.append(str(comment).lstrip("# "))
                lineno -= 1
            elif comment_line > lineno:
                # lexer can get ahead of this line; keep going.
                ...
            else:
                break

        return "\n".join(reversed(results))

    uag_head = transformer.stringify
    uag_body = transformer.pass_through
    uag_user_list = transformer.listify
    uag_user_list_name = transformer.stringify

    def uag(self, uag_token: lark.Token, name, users=None):
        user_group = UserAccessGroup(
            context=context_from_lark_token(self.fn, uag_token),
            name=name,
            users=users or [],
            comments=self.get_recent_comments(uag_token.line),
        )
        self.users[user_group.name] = user_group

    hag_head = transformer.stringify
    hag_body = transformer.pass_through
    hag_host_list = transformer.listify
    hag_host_list_name = transformer.stringify

    def hag(self, hag_token: lark.Token, name, hosts=None):
        host_group = HostAccessGroup(
            context=context_from_lark_token(self.fn, hag_token),
            name=name,
            hosts=hosts or [],
            comments=self.get_recent_comments(hag_token.line),
        )
        self.hosts[host_group.name] = host_group

    def asg(self, asg_token: lark.Token, name, body=None):
        rules, self.rules = self.rules, []
        inputs, self.inputs = self.inputs, {}
        self.groups[str(name)] = AccessSecurityGroup(
            context=context_from_lark_token(self.fn, asg_token),
            comments=self.get_recent_comments(asg_token.line),
            rules=rules,
            name=name,
            inputs=inputs,
        )

    asg_head = transformer.stringify
    asg_body = transformer.stringify
    asg_body_list = transformer.listify
    asg_body_item = transformer.pass_through

    def rule_config(self, rule_token: lark.Token, head, body=None):
        rule_info, self.rule_info = self.rule_info, {}
        token_to_kwarg = {
            "UAG": "users",
            "HAG": "hosts",
            "CALC": "calc",
        }
        kwargs = {
            token_to_kwarg[key]: value
            for key, value in rule_info.items()
        }
        level, options, log_options = head
        if log_options is not None:
            kwargs["log_options"] = log_options

        if self.calc is not None:
            kwargs["calc"] = self.calc[1]
            self.calc = None

        rule = AccessSecurityRule(
            context=context_from_lark_token(self.fn, rule_token),
            comments=self.get_recent_comments(rule_token.line),
            level=level,
            options=options,
            **kwargs,
        )
        self.rules.append(rule)

    rule_body = transformer.pass_through
    rule_list = transformer.listify

    @staticmethod
    def rule_head(mandatory, options):
        level, log_options = mandatory
        return (level, log_options, options)

    @staticmethod
    def rule_head_mandatory(level, options):
        return (int(level), str(options))

    rule_head_options = transformer.pass_through
    rule_log_options = transformer.stringify

    rule_uag_list = transformer.listify_strings
    rule_hag_list = transformer.listify_strings

    def rule_uag(self, uag_token: lark.Token, users):
        self.rule_info[uag_token] = users

    def rule_hag(self, hag_token: lark.Token, hosts):
        self.rule_info[hag_token] = hosts

    def rule_calc(self, calc_token: lark.Token, calc_str):
        self.calc = (calc_token, str(calc_str).strip('" '))

    def inp_config(self, inp_token: lark.Token, link):
        self.inputs[inp_token] = str(link)

    @lark.visitors.v_args(tree=True)
    def asconfig(self, _):
        return AccessSecurityConfig(
            filename=self.fn,
            header=self.get_header_comments(),
            hash=self.contents_hash,
            hosts=self.hosts,
            users=self.users,
            groups=self.groups,
        )


@dataclass
class AccessSecurityConfig:
    """An Access Security Configuration file (ACF) container."""
    filename: Optional[str] = None
    hash: Optional[str] = None
    users: Dict[str, UserAccessGroup] = field(default_factory=dict)
    groups: Dict[str, AccessSecurityGroup] = field(default_factory=dict)
    hosts: Dict[str, HostAccessGroup] = field(default_factory=dict)
    header: str = ""
    comments: List[lark.Token] = field(
        default_factory=list,
        metadata=apischema.metadata.skip,
    )

    def get_group_from_record(self, record: RecordInstance) -> Optional[AccessSecurityGroup]:
        """Get the appropriate access security group for the given record."""
        if record.is_pva:
            return

        return self.groups.get(record.access_security_group, None)

    @classmethod
    def from_string(cls, contents: str, filename: Optional[str] = None) -> AccessSecurityConfig:
        """
        Load access security configuration from a string.

        Parameters
        ----------
        contents : str
            The access security file contents.

        filename : str, optional
            The access security filename to use for LoadContext.
        """
        contents_hash = get_bytes_sha256(contents.encode("utf-8"))
        comments = []

        def add_comment(comment: lark.Token):
            comments.append(
                StringWithContext(
                    str(comment).lstrip("# "),
                    context=context_from_lark_token(filename, comment),
                )
            )

        grammar = lark.Lark.open_from_package(
            "whatrecord",
            "access_security.lark",
            search_paths=("grammar", ),
            parser="lalr",
            lexer_callbacks={"COMMENT": add_comment},
            transformer=_AcfTransformer(filename, contents_hash, comments),
            maybe_placeholders=False,
        )

        return grammar.parse(contents)

    @classmethod
    def from_file_obj(cls, fp, filename: Optional[str] = None) -> AccessSecurityConfig:
        """Load an ACF file from a file object."""
        filename = filename or getattr(fp, "name", str(id(fp)))
        return cls.from_string(fp.read(), filename=filename)

    @classmethod
    def from_file(cls, fn: Union[str, pathlib.Path]) -> AccessSecurityConfig:
        """Load an ACF file from a filename."""
        with open(fn, "rt") as fp:
            return cls.from_file_obj(fp, filename=str(fn))


_handler = ShellStateHandler.generic_handler_decorator


@dataclass
class AccessSecurityState(ShellStateHandler):
    """
    Access Security IOC shell state handler / container.

    Contains hooks for as-related commands and state information.

    Attributes
    ----------
    config : AccessSecurityState
        The access security configuration.
    filename : pathlib.Path
        The access security filename.
    macros : Dict[str, str]
        Macros used when expanding the access security file.
    """
    metadata_key: ClassVar[str] = "asg"
    config: Optional[AccessSecurityConfig] = None
    filename: Optional[pathlib.Path] = None
    macros: Optional[Dict[str, str]] = None

    def post_ioc_init(self):
        super().post_ioc_init()
        if self.filename is None:
            return

        try:
            return {
                "access_security": self._load_access_security()
            }
        except Exception as ex:
            return {
                "access_security": {
                    "exception_class": type(ex).__name__,
                    "error": str(ex),
                }
            }

    @_handler
    def handle_asSetSubstitutions(self, macros: str):
        if self.primary_handler is None:
            return

        macro_context = self.primary_handler.macro_context
        self.macros = macro_context.definitions_to_dict(macros)
        return {
            "macros": self.macros,
            "note": "See iocInit results for details.",
        }

    @_handler
    def handle_asSetFilename(self, filename: str):
        if self.primary_handler is None:
            return

        self.filename = self.primary_handler._fix_path(
            filename
        ).resolve()
        return {
            "filename": str(self.filename),
            "note": "See iocInit results for details.",
        }

    def _load_access_security(self):
        """Load access security settings at iocInit time."""
        if self.primary_handler is None:
            return

        filename, contents = self.primary_handler.load_file(self.filename)
        if self.macros:
            macro_context = MacroContext(use_environment=False)
            macro_context.define(**self.macros)
            contents = "\n".join(
                macro_context.expand(line)
                for line in contents.splitlines()
            )

        self.config = AccessSecurityConfig.from_string(
            contents, filename=str(filename)
        )
        return {
            "filename": filename,
            "macros": self.macros,
        }

    def annotate_record(self, record: RecordInstance) -> Optional[Dict[str, Any]]:
        """Annotate record with access security information."""
        if self.config is not None:
            asg = self.config.get_group_from_record(record)
            if asg is not None:
                return apischema.serialize(asg)



import pytest

from .. import common
from ..common import LoadContext


@pytest.mark.parametrize(
    "ctx, expected",
    [
        pytest.param(
            (LoadContext("a", 0), LoadContext("a", 10), LoadContext("c", 1)),
            (LoadContext("a", 10), LoadContext("c", 1)),
            id="redundant",
        ),
        pytest.param(
            (LoadContext("a", 1), LoadContext("a", 10), LoadContext("c", 1)),
            (LoadContext("a", 1), LoadContext("a", 10), LoadContext("c", 1)),
            id="not-redundant",
        ),
        pytest.param((), (), id="empty"),
    ]
)
def test_redundant_context(ctx, expected):
    assert common.remove_redundant_context(ctx) == expected



"""
SLAC PCDS IocManager configuration file loading support.
"""

import ast
import logging
import os
import pathlib
import re
from typing import Any, Callable, List, Optional, Union

from .common import IocInfoDict
from .util import find_binary_from_hashbang

logger = logging.getLogger(__name__)

KEY_RE = re.compile(r"([a-z_]+)\s*:", re.IGNORECASE)
EPICS_SITE_TOP = os.environ.get("EPICS_SITE_TOP", "/reg/g/pcds/epics")
REQUIRED_KEYS = {"id", "host", "port", "dir", "base_version"}


def validate_config_keys(cfg: IocInfoDict) -> bool:
    """Validate that a configuration has all required keys."""
    return all(key in cfg and cfg[key] for key in REQUIRED_KEYS)


def parse_config(lines: List[str]) -> List[IocInfoDict]:
    """
    Parse an IOC manager config to get its IOCs.

    The parsing here is a bit weak when it comes to syntax errors. This is
    generally not a problem as these files are almost entirely written out by
    the IOC Manager GUI / CLI tools.  Occasionally, a user may edit the file
    manually and introduce syntax errors on certain lines, and this will
    attempt to skip it as best it can.

    Parameters
    ----------
    lines : list of str
        List of raw configuration file lines.

    Returns
    -------
    ioc_info : list of IOC info dictionaries
        List of IOC info
    """
    entries = []
    loading = False
    entry = None

    for line in lines:
        if "procmgr_config" in line:
            loading = True
            continue
        if not loading:
            continue
        if line.lstrip().startswith("#"):
            continue
        if "id:" in line:
            if "}" in line:
                # {id: ... } in a single line
                entries.append(line)
            else:
                # {id: ...
                #  ... }
                entry = line
        elif entry is not None:
            entry += line
            if "}" in entry:
                # {id: ...
                #  ... }   <-- closing line
                entries.append(entry)
                entry = None

    def fix_entry(entry):
        return ast.literal_eval(KEY_RE.sub(r'"\1":', entry.strip(", \t")))

    result = []
    for entry in entries:
        try:
            result.append(fix_entry(entry))
        except Exception:
            logger.error("Failed to fix up IOC manager entry: %s", entry)

    return result


def load_config_file(fn: Union[str, pathlib.Path]) -> List[IocInfoDict]:
    """
    Load a configuration file and return the IOCs it contains.

    Parameters
    ----------
    fn : str or pathlib.Path
        The configuration filename

    Returns
    -------
    ioc_info : list of IOC info dictionaries
        List of IOC info
    """
    with open(fn, "rt") as f:
        lines = f.read().splitlines()

    iocs = parse_config(lines)

    for ioc in list(iocs):
        # For now, assume old database syntax by specifying 3.15:
        ioc["base_version"] = "3.15"
        if not validate_config_keys(ioc):
            iocs.remove(ioc)
        else:
            # Add "config_file" and rename some keys:
            ioc["config_file"] = str(fn)
            ioc["name"] = ioc.pop("id")
            ioc["script"] = find_stcmd(ioc["dir"], ioc["name"])
            ioc["binary"] = find_binary_from_hashbang(ioc["script"])

    return iocs


def find_stcmd(directory: str, ioc_id: str) -> str:
    """Find the startup script st.cmd for a given IOC."""
    if directory.startswith("ioc"):
        directory = os.path.join(EPICS_SITE_TOP, directory)

    suffix = ("iocBoot", ioc_id, "st.cmd")
    # Templated IOCs are... different:
    options = [
        os.path.join(directory, "children", "build", *suffix),
        os.path.join(directory, "build", *suffix),
        os.path.join(directory, *suffix),
        os.path.join(directory, "st.cmd"),
    ]

    for option in options:
        if os.path.exists(option):
            return option

    # Guess at what's correct:
    return options[-1]


def get_iocs_from_configs(
    configs: List[Union[str, pathlib.Path]],
    sorter: Optional[Callable[[IocInfoDict], Any]] = None
) -> List[IocInfoDict]:
    """
    Get IOC information in a list of dictionaries.

    Parameters
    ----------
    configs : list[str or pathlib.Path]
        Configuration filenames to load.
    sorter : callable, optional
        Sort IOCs with this, defaults to sorting by host name and then IOC name.

    Returns
    -------
    ioc_info : list of IOC info dictionaries
        List of IOC info
    """
    configs = [
        pathlib.Path(config).resolve()
        for config in configs or []
    ]

    def default_sorter(ioc):
        return (ioc["host"], ioc["name"])

    iocs = (
        ioc
        for fn in set(configs)
        for ioc in load_config_file(fn)
    )

    return list(sorted(iocs, key=sorter or default_sorter))



import dataclasses
import logging
import pathlib
from dataclasses import field
from typing import Dict, List, Union

from .common import IocInfoDict, IocMetadata
from .util import run_script_with_json_output

logger = logging.getLogger(__name__)


class _IocInfoFinder:
    """
    An IOC information "finder" base class.

    Subclasses of this support different ways of finding IOCs from
    user-specified parameters.  See further information in subclasses.
    """
    scripts: Dict[pathlib.Path, IocMetadata]

    def add_or_update_entry(self, info: IocMetadata):
        self.scripts[info.script] = info

    async def update(self):
        """
        Subclasses must implement this method.

        The backend server will call this periodically based on the
        user-configured settings.
        """
        ...


@dataclasses.dataclass
class IocScriptExternalLoader(_IocInfoFinder):
    """
    An IOC information "finder" that utilizes an external executable script.

    That script must return a JSON object that is compatible with
    `IocMetadata.from_dict`.
    """
    #: The script to run.
    load_script: str
    scripts: Dict[pathlib.Path, IocMetadata] = field(default_factory=dict)

    def __post_init__(self):
        if not self.load_script:
            raise ValueError("load_script is required")

    async def update(self):
        result = await run_script_with_json_output(self.load_script)
        for info in (result or {}):
            self.add_or_update_entry(IocMetadata.from_dict(info))
        return await super().update()


@dataclasses.dataclass
class IocScriptStaticInfoList(_IocInfoFinder):
    """
    An IOC information "finder" that is provided a list of IOC information
    dictionaries.

    This is mainly LCLS-specific as it integrates with their "IOC Manager"
    utility.
    """
    ioc_infos: List[IocInfoDict]
    scripts: Dict[pathlib.Path, IocMetadata] = field(default_factory=dict)

    async def update(self):
        if self.scripts:
            # The list is static; there wil be no updates
            return

        for info in self.ioc_infos:
            self.add_or_update_entry(IocMetadata.from_dict(info))


@dataclasses.dataclass
class IocScriptStaticList(_IocInfoFinder):
    """
    An IOC finder that does not do much finding; it expects a static list of
    scripts to load at startup.
    """
    script_list: List[Union[str, pathlib.Path]]
    scripts: Dict[pathlib.Path, IocMetadata] = field(default_factory=dict)

    async def update(self):
        if self.scripts:
            # The list is static; there wil be no updates
            return

        for fn in self.script_list:
            self.add_or_update_entry(IocMetadata.from_file(fn))



from __future__ import annotations

import datetime
import pathlib
import re
from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, List, Optional, Union

import lark

from . import settings, transformer
from .common import AnyPath, FullLoadContext, ShellStateHandler
from .db import RecordInstance
from .macro import MacroContext, macros_from_string
from .transformer import context_from_token


@dataclass
class RestoreError:
    context: FullLoadContext
    number: int
    description: str


@dataclass
class RestoreValue:
    context: FullLoadContext
    pvname: str
    record: str
    field: str
    value: Union[str, List[str]]


@dataclass
class AutosaveRestoreFile:
    """Representation of an autosave restore (.sav) file."""
    filename: str
    values: Dict[str, Dict[str, RestoreValue]] = field(default_factory=dict)
    disconnected: List[str] = field(default_factory=list)
    errors: List[RestoreError] = field(default_factory=list)
    comments: List[str] = field(default_factory=list)

    @classmethod
    def from_string(
        cls,
        contents: str,
        filename: AnyPath = "",
        macros: Optional[Dict[str, str]] = None
    ) -> AutosaveRestoreFile:
        """Load an autosave file given its string contents."""
        grammar = lark.Lark.open_from_package(
            "whatrecord",
            "autosave_save.lark",
            search_paths=("grammar", ),
            parser="earley",
            maybe_placeholders=False,
            propagate_positions=True,
        )

        if macros:
            contents = MacroContext(macros=macros).expand_by_line(contents)

        return _AutosaveRestoreTransformer(cls, filename).transform(
            grammar.parse(contents)
        )

    @classmethod
    def from_file_obj(
        cls, fp, filename: AnyPath, macros: Optional[Dict[str, str]] = None
    ) -> AutosaveRestoreFile:
        """Load an autosave file given a file object."""
        return cls.from_string(
            fp.read(),
            filename=getattr(fp, "name", filename),
            macros=macros,
        )

    @classmethod
    def from_file(
        cls, filename: AnyPath, macros: Optional[Dict[str, str]] = None
    ) -> AutosaveRestoreFile:
        """
        Load an autosave restore (.sav) file.

        Parameters
        ----------
        filename : pathlib.Path or str
            The filename.

        Returns
        -------
        file : AutosaveRestoreFile
            The resulting parsed file.
        """
        with open(filename, "rt") as fp:
            return cls.from_string(fp.read(), filename=filename, macros=macros)


@lark.visitors.v_args(inline=True)
class _AutosaveRestoreTransformer(lark.visitors.Transformer):
    def __init__(self, cls, fn, visit_tokens=False):
        super().__init__(visit_tokens=visit_tokens)
        self.fn = str(fn)
        self.cls = cls
        self.errors = []
        self.restores = {}
        self.comments = []
        self.disconnected = []

    @lark.visitors.v_args(tree=True)
    def restore(self, body):
        return self.cls(
            filename=self.fn,
            values=self.restores,
            errors=self.errors,
            comments=self.comments,
            disconnected=self.disconnected,
        )

    restore_item = transformer.pass_through

    def error(self, exclamation, error_number, error_description=None):
        self.errors.append(
            RestoreError(
                context=transformer.context_from_token(self.fn, exclamation),
                number=error_number,
                description=str(error_description or ""),
            )
        )

    def value_restore(self, pvname, value=None):
        context = context_from_token(self.fn, pvname)
        if "." in pvname:
            record, field = pvname.split(".", 1)
        else:
            record, field = pvname, "VAL"

        record = str(record)
        if record not in self.restores:
            self.restores[record] = {}

        self.restores[record][field.lstrip(".")] = RestoreValue(
            context=context,
            record=record,
            pvname=f"{record}.{field}" if field else record,
            field=str(field),
            value=value if isinstance(value, list) else _fix_value(value),
        )

    value = transformer.pass_through
    record_name = transformer.pass_through
    scalar_value = transformer.pass_through
    pvname = transformer.pass_through

    def array_elements(self, *elements):
        return [_fix_value(elem) for elem in elements]

    def array_value(self, array_marker, array_begin, elements, array_end):
        return list(elements)

    def error_number(self, number):
        return int(number)

    error_description = transformer.stringify

    def comment(self, comment):
        comment = comment.lstrip("# ")
        if not comment:
            return

        parts = comment.split(" ")
        if len(parts) == 3 and parts[-2:] == ["Search", "Issued"]:
            self.disconnected.append(parts[0])
        else:
            self.comments.append(comment)


def _strip_double_quote(value: str) -> str:
    """Strip one leading/single trailing double-quote."""
    if value[0] == '"':
        value = value[1:]
    if value[-1] == '"':
        value = value[:-1]
    return value


RE_REMOVE_ESCAPE = re.compile(r"\\(.)")


def _fix_value(value: Optional[str]) -> str:
    """Remove quotes, and fix up escaping."""
    if value is None:
        # Value can be empty in autosave files
        return ""
    value = _strip_double_quote(value)
    return RE_REMOVE_ESCAPE.sub(r"\1", value)


@dataclass
class AutosaveSet:
    context: FullLoadContext
    request_filename: str
    save_filename: str
    period: Optional[int] = None
    trigger_channel: Optional[str] = None
    macros: Dict[str, str] = field(default_factory=dict)
    method: str = "manual"


@dataclass
class AutosaveRestorePassFile:
    context: FullLoadContext
    save_filename: str
    macros: Dict[str, str] = field(default_factory=dict)
    pass_number: int = 0

    load_timestamp: Optional[datetime.datetime] = None
    file_timestamp: Optional[datetime.datetime] = None
    data: Optional[AutosaveRestoreFile] = None

    def update(self, save_path: pathlib.Path) -> AutosaveRestoreFile:
        """Update the autosave .sav file from disk."""
        fn = save_path / self.save_filename
        file_timestamp = datetime.datetime.fromtimestamp(fn.stat().st_mtime)
        if self.file_timestamp is not None and file_timestamp == self.file_timestamp:
            if self.data is not None:
                return self.data

        if self.load_timestamp is not None and self.data is not None:
            dt = datetime.datetime.now() - self.load_timestamp
            if dt.total_seconds() < settings.AUTOSAVE_RELOAD_PERIOD:
                return self.data

        self.file_timestamp = file_timestamp
        self.load_timestamp = datetime.datetime.now()
        self.data = AutosaveRestoreFile.from_file(
            fn,
            macros=self.macros,
        )
        return self.data


_handler = ShellStateHandler.generic_handler_decorator


@dataclass
class AutosaveState(ShellStateHandler):
    """The state of autosave in an IOC."""

    metadata_key: ClassVar[str] = "autosave"

    configured: bool = False
    request_paths: List[pathlib.Path] = field(default_factory=list)
    save_path: pathlib.Path = field(default_factory=pathlib.Path)
    sets: Dict[str, AutosaveSet] = field(default_factory=dict)
    restore_files: Dict[str, AutosaveRestorePassFile] = field(default_factory=dict)
    incomplete_sets_ok: Optional[bool] = None  # default: True
    dated_backups: Optional[bool] = None  # default: True
    date_period_minutes: Optional[int] = None  # default: 0
    num_seq_files: Optional[int] = None  # default: 3
    seq_period: Optional[int] = None  # default: 0
    retry_seconds: Optional[int] = None  # default: 0
    ca_reconnect: Optional[bool] = None  # default: False
    callback_timeout: Optional[int] = None  # default: 0
    task_priority: Optional[int] = None  # default: 0
    nfs_host: Optional[str] = None
    use_status_pvs: Optional[bool] = None  # default: False
    status_prefix: Optional[str] = None
    file_permissions: Optional[int] = None  # default: 0o664
    debug: Optional[int] = None  # default: 0

    @property
    def save_name_pv(self) -> Optional[str]:
        """The save name PV, derived from the macro context."""
        if self.primary_handler is not None:
            return self.primary_handler.macro_context.get("SAVENAMEPV")

    @property
    def save_path_pv(self) -> Optional[str]:
        """The save path PV, derived from the macro context."""
        if self.primary_handler is not None:
            return self.primary_handler.macro_context.get("SAVEPATHPV")

    # save_restore.c

    @_handler
    def handle_fdbrestore(self, filename: str = ""):
        """
        If save_file refers to a save set that exists in memory, then PV's in
        the save set will be restored from values in memory. Otherwise, this
        functions restores the PV's in <saveRestorePath>/<save_file> and
        creates a new backup file "<saveRestorePath>/<save_file>.bu". The
        effect probably will not be the same as a boot-time restore, because
        caput() calls are used instead of static database access dbPutX()
        calls. Record processing will result from caput()'s to inherently
        process- passive fields.
        """

    @_handler
    def handle_fdbrestoreX(self, filename="", macrostring=""):
        """
        This function restores from the file <saveRestorePath>/<save_file>,
        which can look just like a save file, but which needn't end with <END>.
        No backup file will be written. The effect probably will not be the
        same as a boot-time restore, because caput() calls are used instead of
        static database access dbPut*() calls. Record processing will result
        from caput()'s to inherently process-passive fields.
        """

    @_handler
    def handle_manual_save(self, request_file: str = ""):
        """
        Cause current PV values for the request file to be saved. Any request
        file named in a create_xxx_set() command can be saved manually.
        """

    @_handler
    def handle_set_savefile_name(self, request_file: str = "", save_filename: str = ""):
        """
        If a save set has already been created for the request file, this
        function will change the save file name.
        """
        try:
            set_ = self.sets[request_file]
        except KeyError:
            raise ValueError("Request file not configured")

        set_.save_filename = save_filename

    @_handler
    def handle_create_periodic_set(
        self, filename: str = "", period: int = 0, macro_string: str = "", *_
    ):
        """
        Create a save set for the request file. The save file will be written
        every period seconds.
        """
        self.sets[filename] = AutosaveSet(
            context=self.get_load_context(),
            request_filename=filename,
            save_filename="{}.sav".format(pathlib.Path(filename).stem),
            period=period,
            trigger_channel=None,
            macros=macros_from_string(macro_string),
            method="periodic",
        )

    @_handler
    def handle_create_triggered_set(
        self, filename: str = "", trigger_channel: str = "", macro_string: str = "", *_
    ):
        """
        Create a save set for the request file. The save file will be written
        whenever the PV specified by trigger_channel is posted. Normally this
        occurs when the PV's value changes.
        """
        self.sets[filename] = AutosaveSet(
            context=self.get_load_context(),
            request_filename=filename,
            save_filename="{}.sav".format(pathlib.Path(filename).stem),
            period=None,
            trigger_channel=trigger_channel,
            macros=macros_from_string(macro_string),
            method="triggered",
        )

    @_handler
    def handle_create_monitor_set(
        self, filename: Optional[str] = None, period: int = 0, macro_string: str = "", *_
    ):
        """
        Create a save set for the request file. The save file will be written
        every period seconds, if any PV in the save set was posted
        (changed value) since the last write.
        """
        if filename is None:
            # An indicator to "start the save task"
            return

        self.sets[filename] = AutosaveSet(
            context=self.get_load_context(),
            request_filename=filename,
            save_filename="{}.sav".format(pathlib.Path(filename).stem),
            period=int(period),
            trigger_channel=None,
            macros=macros_from_string(macro_string),
            method="monitor",
        )

    @_handler
    def handle_create_manual_set(self, filename: str = "", macro_string: str = ""):
        """
        Create a save set for the request file. The save file will be written
        when the function manual_save() is called with the same request-file
        name.
        """
        self.sets[filename] = AutosaveSet(
            context=self.get_load_context(),
            request_filename=filename,
            save_filename="{}.sav".format(pathlib.Path(filename).stem),
            period=None,
            trigger_channel=None,
            macros=macros_from_string(macro_string),
            method="manual",
        )

    @_handler
    def handle_save_restoreShow(self, verbose: int = 0):
        """Show the save restore status."""
        return self.sets

    @_handler
    def handle_set_requestfile_path(self, path: str = "", subpath: str = ""):
        full_path = (pathlib.Path(path) / subpath).resolve()
        if full_path not in self.request_paths:
            self.request_paths.append(full_path)

    @_handler
    def handle_set_savefile_path(self, path: str = "", subpath: str = ""):
        """
        Called before iocInit(), this function specifies the path to be
        prepended to save-file and restore-file names. pathsub, if present,
        will be appended to path, if present, with a separating '/', whether or
        not path ends or pathsub begins with '/'. If the result does not end in
        '/', one will be appended to it.

        If save_restore is managing its own NFS mount, this function specifies
        the mount point, and calling it will result in an NFS mount if all
        other requirements have already been met. If a valid NFS mount already
        exists, the file system will be dismounted and then mounted with the
        new path name. This function can be called at any time.
        """
        path = pathlib.Path(path) / subpath
        if self.primary_handler is not None:
            path = self.primary_handler._fix_path(path)
        self.save_path = path.resolve()

    @_handler
    def handle_set_saveTask_priority(self, priority: int = 0):
        """Set the priority of the save_restore task."""
        self.task_priority = int(priority)

    @_handler
    def handle_save_restoreSet_NFSHost(
        self, hostname: str = "", address: str = "", mntpoint: str = "", *_
    ):
        """
        Specifies the name and IP address of the NFS host. If both have been
        specified, and set_savefile_path() has been called to specify the file
        path, save_restore will manage its own NFS mount. This allows
        save_restore to recover from a reboot of the NFS host
        (that is, a stale file handle) and from some kinds of tampering with
        the save_restore directory.
        """
        self.nfs_host = f"nfs://{hostname}/{mntpoint} ({address})"

    @_handler
    def handle_remove_data_set(self, filename: str = ""):
        """If a save set has been created for request_file, this function will delete it."""
        ...

    @_handler
    def handle_reload_periodic_set(
        self, filename: str = "", period: int = 0, macro_string: str = "", *_
    ):
        """
        This function allows you to change the PV's and the period associated
        with a save set created by create_periodic_set().
        """

    @_handler
    def handle_reload_triggered_set(
        self, filename: str = "", trigger_channel: str = "", macro_string: str = "", *_
    ):
        """
        This function allows you to change the PV's and the trigger channel
        associated with a save set created by create_triggered_set().
        """

    @_handler
    def handle_reload_monitor_set(
        self, filename: str = "", period: int = 0, macro_string: str = "", *_
    ):
        """
        This function allows you to change the PV's and the period associated
        with a save set created by create_monitor_set().
        """

    @_handler
    def handle_reload_manual_set(self, filename: str = "", macrostring: str = ""):
        """
        This function allows you to change the PV's associated with a save set
        created by create_manual_set().
        """

    @_handler
    def handle_save_restoreSet_Debug(self, level: int = 0):
        """
        Sets the value (int) save_restoreDebug (initially 0). Increase to get
        more informational messages printed to the console.
        """
        self.debug = int(level)

    @_handler
    def handle_save_restoreSet_NumSeqFiles(self, numSeqFiles: int = 0):
        """
        Sets the value of (int) save_restoreNumSeqFiles (initially 3). This is
        the number of sequenced backup files to be maintained.
        """
        self.num_seq_files = int(numSeqFiles)
        if not (0 <= self.num_seq_files <= 10):
            raise ValueError("numSeqFiles must be between 0 and 10 inclusive.")

    @_handler
    def handle_save_restoreSet_SeqPeriodInSeconds(self, period: int = 0):
        """
        Sets the value of (int) save_restoreSeqPeriodInSeconds (initially 60).
        Sequenced backup files will be written with this period.
        """
        self.seq_period = int(period)
        if self.seq_period < 10:
            raise ValueError("period must be 10 or greater.")

    @_handler
    def handle_save_restoreSet_IncompleteSetsOk(self, ok: int = 0):
        """
        Sets the value of (int) save_restoreIncompleteSetsOk (initially 1). If
        set to zero, save files will not be restored at boot time unless they
        are perfect, and they will not be overwritten at save time unless a
        valid CA connection and value exists for every PV in the list.
        """
        self.incomplete_sets_ok = bool(ok)

    @_handler
    def handle_save_restoreSet_DatedBackupFiles(self, ok: int = 0):
        """
        Sets the value of (int) save_restoreDatedBackupFiles (initially 1). If
        zero, the backup file written at reboot time
        (a copy of the file from which PV values are restored) will have the
        suffix '.bu', and will be overwritten every reboot. If nonzero, each
        reboot will leave behind its own backup file.
        """
        self.dated_backups = bool(ok)

    @_handler
    def handle_save_restoreSet_status_prefix(self, prefix: str = ""):
        """
        Specifies the prefix to be used to construct the names of PV's with
        which save_restore reports its status. If you want autosave to update
        status PVs as it operates, you must call this function and load the
        database save_restoreStatus.db, specifying the same prefix in both
        commands.
        """
        self.status_prefix = prefix

    @_handler
    def handle_save_restoreSet_FilePermissions(self, permissions: int = 0):
        """
        Specify the file permissions used to create new .sav files. This
        integer value will be supplied, exactly as given, to the system call,
        open(), and to the call fchmod(). Typically, file permissions are set
        with an octal number, such as 0640, and
        save_restoreSet_FilePermissions() will confirm any number given to it
        by echoing it to the console as an octal number.
        """
        self.file_permissions = int(permissions)

    @_handler
    def handle_save_restoreSet_RetrySeconds(self, seconds: int = 0):
        """
        Specify the time delay between a failed .sav-file write and the retry
        of that write. The default delay is 60 seconds. If list-PV's change
        during the delay, the new values will be written.
        """
        self.retry_seconds = int(seconds)

    @_handler
    def handle_save_restoreSet_UseStatusPVs(self, ok: int = 0):
        """
        Specifies whether save_restore should report its status to a preloaded
        set of EPICS PV's (contained in the database save_restoreStatus.db). If
        the argument is '0', then status PV's will not be used.
        """
        self.use_status_pvs = bool(ok)

    @_handler
    def handle_save_restoreSet_CAReconnect(self, ok: int = 0):
        """
        Specify whether autosave should periodically retry connecting to PVs
        whose initial connection attempt failed. Currently, the
        connection-retry interval is hard-wired at 60 seconds.
        """
        self.ca_reconnect = bool(ok)

    @_handler
    def handle_save_restoreSet_CallbackTimeout(self, timeout: int = 0):
        """
        Specify the time interval in seconds between forced save-file writes.
        (-1 means forever). This is intended to get save files written even if
        the normal trigger mechanism is broken.
        """
        self.callback_timeout = int(timeout)

    @_handler
    def handle_asVerify(
        self, filename: str = "", verbose: int = 0, restoreFileName: str = "", *_
    ):
        """
        Compare PV values in the IOC with values written in filename
        (which should be an autosave restore file, or at least look like one).
        If restoreFileName is not empty, write a new restore file.
        """
        ...

    @_handler
    def handle_save_restoreSet_periodicDatedBackups(self, periodMinutes: int = 0):
        """
        Sets the value of (int) save_restoreDatedBackupFiles (initially 1). If
        zero, the backup file written at reboot time
        (a copy of the file from which PV values are restored) will have the
        suffix '.bu', and will be overwritten every reboot. If nonzero, each
        reboot will leave behind its own backup file.
        """
        self.dated_backups = True
        self.date_period_minutes = int(periodMinutes)

    # dbrestore.c

    @_handler
    def handle_set_pass0_restoreFile(self, file: str = "", macro_string: str = ""):
        """
        This function specifies a save file to be restored during iocInit,
        before record initialization. An unlimited number of files can be
        specified using calls to this function. If the file name begins with
        "/", autosave will use it as specified; otherwise, autosave will
        prepend the file path specified to set_savefile_path(). The second
        argument is optional.
        """
        self.restore_files[file] = AutosaveRestorePassFile(
            context=self.get_load_context(),
            save_filename=file,
            macros=macros_from_string(macro_string),
            pass_number=0,
        )
        return {
            "autosave": f"Added pass 0 restore file {self.save_path}/{file}"
        }

    @_handler
    def handle_set_pass1_restoreFile(self, file: str = "", macro_string: str = ""):
        """
        This function specifies a save file to be restored during iocInit,
        after record initialization. An unlimited number of files can be
        specified using calls to this function. If the file name begins with
        "/", autosave will use it as specified; otherwise, autosave will
        prepend the file path specified to set_savefile_path(). The second
        argument is optional.
        """
        self.restore_files[file] = AutosaveRestorePassFile(
            context=self.get_load_context(),
            save_filename=file,
            macros=macros_from_string(macro_string),
            pass_number=1,
        )
        return {
            "autosave": f"Added pass 1 restore file {self.save_path}/{file}"
        }

    @_handler
    def handle_dbrestoreShow(self):
        """
        List all the save sets currently being managed by the save_restore
        task. If (verbose != 0), lists the PV's as well.
        """
        ...

    @_handler
    def handle_makeAutosaveFileFromDbInfo(
        self, filename: str = "", info_name: str = "", *_
    ):
        """
        Search through the EPICS database
        (that is, all EPICS records loaded into an IOC) for 'info' nodes named
        info_name; construct a list of PV names from the associated info_values
        found, and write the PV names to the file fileBaseName. If fileBaseName
        does not contain the string '.req', this string will be appended to it.
        See makeAutosaveFiles() for more information.
        """

    @_handler
    def handle_makeAutosaveFiles(self):
        """
        Search through the EPICS database
        (that is, all EPICS records loaded into an IOC) for info nodes named
        'autosaveFields' and 'autosaveFields_pass0'; construct lists of PV
        names from the associated info values, and write the PV names to the
        files 'info_settings.req' and 'info_positions.req', respectively.
        """

    @_handler
    def handle_eraseFile(self, filename: str = ""):
        """Erase (empty) an autosave file."""
        ...

    @_handler
    def handle_appendToFile(self, filename: str = "", line: str = ""):
        """
        Append line to a file.

        For example, to add a line to built_settings.req yourself:

        appendToFile("built_settings.req", '$(P)userStringSeqEnable')
        """
        ...

    @_handler
    def handle_autosaveBuild(
        self, filename: str = "", reqFileSuffix: str = "", on: int = 0, *_
    ):
        """
        It's tedious and error prone to have these entries separately
        maintained, so autosave can do the request-file part for you. To do
        this, you tell autosave to arrange to be called whenever
        dbLoadRecords() is called
        (note that dbLoadTemplate() calls dbLoadRecords()), you tell it how to
        make a request-file name from a database-file name, and you give it the
        name of the request file you want it to build. You can do this with the
        following command:

            autosaveBuild("built_settings.req", "_settings.req", 1)

        This tells autosave to do the following:

            1. Begin building the file built_settings.req. If this is the first
               call that mentions built_settings.req, erase the file.
            2. Generate request-file names by stripping ".db", or ".vdb", or
               ".template" from database-file names, and adding the suffix
               "_settings.req".
            3. Enable (disable) automated building if the third argument is 1
               (0).

        While automated building is enabled, autosave will generate
        request-file names and search for those files in its request-file path.
        If it finds a request file, it will add the appropriate line to
        built_settings.req.

        All this does is get the file built_settings.req built. If you want it
        to be used, you must add the following line to auto_settings.req:

            file built_settings.req P=$(P)
        """

    def annotate_record(self, record: RecordInstance) -> Optional[Dict[str, Any]]:
        metadata = {}
        for restore_file in self.restore_files.values():
            try:
                data = restore_file.update(self.save_path)
            except FileNotFoundError:
                fn = self.save_path / restore_file.save_filename
                metadata.setdefault("error", []).append(
                    f"Restore file not found: {fn}"
                )
                continue

            record_data = data.values.get(record.name, None)
            if record_data is not None:
                metadata.setdefault("restore", []).append(record_data)
            for pvname in data.disconnected:
                if pvname.split(".")[0] == record.name:
                    if "." in pvname:
                        field = pvname.split(".", 1)[1]
                    else:
                        field = "VAL"
                    metadata.setdefault("disconnected", []).append(
                        field
                    )

        return metadata if metadata else None



"""
The following is borrowed from pygraphviz, reimplemented to allow for asyncio
compatibility.

Some form of it may move upstream at some point.  Requires graphviz<0.18 due
to API changes.
"""

import asyncio
import logging
import os

import graphviz as gv

logger = logging.getLogger(__name__)


async def async_render(
    engine, format, filepath, renderer=None, formatter=None, quiet=False
):
    """
    Async Render file with Graphviz ``engine`` into ``format``,  return result
    filename.

    Parameters
    ----------
    engine :
        The layout commmand used for rendering (``'dot'``, ``'neato'``, ...).
    format :
        The output format used for rendering (``'pdf'``, ``'png'``, ...).
    filepath :
        Path to the DOT source file to render.
    renderer :
        The output renderer used for rendering (``'cairo'``, ``'gd'``, ...).
    formatter :
        The output formatter used for rendering (``'cairo'``, ``'gd'``, ...).
    quiet : bool
        Suppress ``stderr`` output from the layout subprocess.

    Returns
    -------
    The (possibly relative) path of the rendered file.

    Raises
    ------
    ValueError
        If ``engine``, ``format``, ``renderer``, or ``formatter`` are not
        known.

    graphviz.RequiredArgumentError
        If ``formatter`` is given but ``renderer`` is None.

    graphviz.ExecutableNotFound
        If the Graphviz executable is not found.

    subprocess.CalledProcessError
        If the exit status is non-zero.

    Notes
    -----
    The layout command is started from the directory of ``filepath``, so that
    references to external files (e.g. ``[image=...]``) can be given as paths
    relative to the DOT source file.
    """
    # Adapted from graphviz under the MIT License (MIT) Copyright (c) 2013-2020
    # Sebastian Bank
    dirname, filename = os.path.split(filepath)
    cmd, rendered = gv.backend.command(engine, format, filename, renderer, formatter)

    if dirname:
        cwd = dirname
        rendered = os.path.join(dirname, rendered)
    else:
        cwd = None

    proc = await asyncio.create_subprocess_exec(
        *cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        cwd=cwd,
    )

    (stdout, stderr) = await proc.communicate()
    if proc.returncode:
        raise gv.backend.CalledProcessError(
            proc.returncode, cmd, output=stdout, stderr=stderr
        )
    return rendered


class AsyncDigraph(gv.Digraph):
    async def async_render(
        self,
        filename=None,
        directory=None,
        view=False,
        cleanup=False,
        format=None,
        renderer=None,
        formatter=None,
        quiet=False,
        quiet_view=False,
    ):
        """
        Save the source to file and render with the Graphviz engine.

        Parameters
        ----------
        filename :
            Filename for saving the source (defaults to ``name`` + ``'.gv'``)
        directory :
            (Sub)directory for source saving and rendering.
        view (bool) :
            Open the rendered result with the default application.
        cleanup (bool) :
            Delete the source file after rendering.
        format :
            The output format used for rendering (``'pdf'``, ``'png'``, etc.).
        renderer :
            The output renderer used for rendering (``'cairo'``, ``'gd'``, ...).
        formatter :
            The output formatter used for rendering (``'cairo'``, ``'gd'``, ...).
        quiet (bool) :
            Suppress ``stderr`` output from the layout subprocess.
        quiet_view (bool) :
            Suppress ``stderr`` output from the viewer process;
           implies ``view=True``, ineffective on Windows.
        Returns:
            The (possibly relative) path of the rendered file.

        Raises
        ------
        ValueError
            If ``format``, ``renderer``, or ``formatter`` are not known.

        graphviz.RequiredArgumentError
            If ``formatter`` is given but ``renderer`` is None.

        graphviz.ExecutableNotFound
            If the Graphviz executable is not found.

        subprocess.CalledProcessError
            If the exit status is non-zero.

        RuntimeError
            If viewer opening is requested but not supported.

        Notes
        -----
        The layout command is started from the directory of ``filepath``, so that
        references to external files (e.g. ``[image=...]``) can be given as paths
        relative to the DOT source file.
        """
        # Adapted from graphviz under the MIT License (MIT) Copyright (c) 2013-2020
        # Sebastian Bank
        filepath = self.save(filename, directory)

        if format is None:
            format = self._format

        rendered = await async_render(
            self._engine,
            format,
            filepath,
            renderer=renderer,
            formatter=formatter,
            quiet=quiet,
        )

        if cleanup:
            logger.debug("delete %r", filepath)
            os.remove(filepath)

        if quiet_view or view:
            self._view(rendered, self._format, quiet_view)

        return rendered



import asyncio
import collections
import fnmatch
import functools
import gzip
import json
import logging
import os
import pathlib
import re
import tempfile
import time
from typing import Any, Dict, List, Optional, Tuple, Union

import apischema
import graphviz
from aiohttp import web

from .. import common, gateway, graph, ioc_finder, settings
from ..common import (AnyPath, LoadContext, RecordInstance, StringWithContext,
                      WhatRecord)
from ..shell import (LoadedIoc, ScriptContainer, ShellState,
                     load_startup_scripts_with_metadata)
from .common import (FileInfoRawResponse, FileInfoResponse,
                     IocGetDuplicatesResponse, IocGetMatchesResponse,
                     IocGetMatchingRecordsResponse, IocMetadata, PluginResults,
                     PVGetInfo, PVGetMatchesResponse, PVRelationshipResponse,
                     PVShortRelationshipResponse, ServerPluginSpec,
                     TooManyRecordsError)
from .util import TaskHandler

try:
    import tracemalloc
except ImportError:
    # tracemalloc unavailable on pypy
    tracemalloc = None

TRUE_VALUES = {"1", "true", "True"}

logger = logging.getLogger(__name__)
_log_handler = None
tracemalloc_snapshot = None


def serialized_response(obj: Any) -> web.Response:
    """Return an apischema-serialized JSON response of a dataclass instance."""
    return web.json_response(apischema.serialize(obj))


class ServerLogHandler(logging.Handler):
    def __init__(self, message_count: int = 1000, level="DEBUG"):
        super().__init__(level=level)
        self.formatter = logging.Formatter(
            "%(asctime)s - PID %(process)d %(filename)18s: %(lineno)-3s "
            "%(funcName)-18s %(levelname)-8s %(message)s"
        )
        self.message_count = message_count
        self.messages = collections.deque(maxlen=message_count)

    def emit(self, record):
        self.messages.append(self.format(record))


def compile_patterns(patterns: Tuple[str, ...], flags=re.IGNORECASE, use_regex=False):
    """Compile regular expression (or glob) patterns with `re.compile`."""
    if use_regex:
        return re.compile("|".join(patterns), flags=flags)

    return re.compile(
        "|".join(fnmatch.translate(pattern) for pattern in patterns),
        flags=flags,
    )


def get_patterns(
    query, key: str = "pattern", regex_key: str = "regex"
) -> Tuple[bool, Tuple[str, ...]]:
    """Get glob/regex patterns from a server query."""
    use_regex = query.get(regex_key, "false").lower() in TRUE_VALUES
    default = (".*" if use_regex else "*",)
    return use_regex, tuple(query.getall(key, default))


class ServerState:
    running: bool
    container: ScriptContainer
    gateway_config: gateway.GatewayConfig
    script_loaders: List[ioc_finder._IocInfoFinder]
    ioc_metadata: List[IocMetadata]
    _update_count: int

    def __init__(
        self,
        startup_scripts: Optional[List[str]] = None,
        script_loaders: Optional[List[str]] = None,
        standin_directories: Optional[Dict[str, str]] = None,
        gateway_config: Optional[str] = None,
        plugins: Optional[List[ServerPluginSpec]] = None,
    ):
        self.running = False
        self.container = ScriptContainer()
        self.gateway_config = None
        self.gateway_config_path = gateway_config
        self.ioc_metadata = []
        self.plugins = plugins or []
        self.plugins_by_name = {
            plugin.name: plugin
            for plugin in plugins or []
        }
        self.script_relations = {}
        self.standin_directories = standin_directories or {}
        self.tasks = TaskHandler()
        self.script_loaders = [
            ioc_finder.IocScriptStaticList(startup_scripts or [])
        ] + [
            ioc_finder.IocScriptExternalLoader(loader)
            for loader in script_loaders or []
        ]
        self._update_count = 0

    async def trigger_manual_update(self, annotate_records: bool = True) -> None:
        """Manual re-scan of everything."""
        updated = await self._update_script_step()
        logger.debug(
            "Manual update trigger. Updates %s",
            "found" if updated else "not found",
        )
        for plugin in self.plugins:
            await self._update_plugin(plugin)

        # self.container.pv_relations = graph.build_database_relations(
        #     database=self.container.database,
        #     record_types=self.container.record_types,  # <-- inaccurate?
        #     aliases=self.container.aliases,
        # )
        self.script_relations = graph.build_script_relations(
            self.container.database,
            self.container.pv_relations,
        )

        if not annotate_records:
            return

        for ioc in self.container.scripts.values():
            logger.info(
                "Annotating %d records for IOC %s",
                len(ioc.shell_state.database), ioc.name,
            )
            for record in ioc.shell_state.database.values():
                record.metadata.update(self.get_record_annnotations(ioc, record))
                for fn, hash_ in ioc.shell_state.loaded_files.items():
                    if fn not in self.container.loaded_files:
                        logger.info(
                            "Record annotation for %s loaded a new file dynamically: %s (%s)",
                            ioc.name,
                            fn,
                            hash_,
                        )
                        self.container.loaded_files[fn] = hash_

    def get_file_info(self, filename: AnyPath, recorded_hash: str = "") -> FileInfoResponse:
        # TODO recorded hash may differ from what's on disk!
        ioc_name = self.container.startup_script_to_ioc.get(str(filename), None)
        if ioc_name:
            logger.debug("File %s is from IOC %s", filename, ioc_name)
            loaded_ioc = self.container.scripts[ioc_name]
            script_info = loaded_ioc.script
            ioc_md = loaded_ioc.metadata
            return FileInfoResponse(
                script=script_info,
                ioc=ioc_md,
            )

        try:
            with open(filename, "rt") as fp:
                lines = fp.read().splitlines()
        except Exception as ex:
            lines = [
                f"whatrecord unable to dump file due to "
                f"{ex.__class__.__name__}: {ex}"
            ]

        return FileInfoResponse(
            script=None,
            ioc=None,
            raw=FileInfoRawResponse(
                lines=lines,
                path=str(filename),
                hash=recorded_hash,
            ),
        )

    def dump_iocs(self) -> dict:
        return {
            ioc.metadata.name: {
                "md": apischema.serialize(IocMetadata, ioc.metadata),
                "ioc": apischema.serialize(LoadedIoc, ioc),
            }
            for ioc in self.container.scripts.values()
            if ioc.metadata and ioc.metadata.name
        }

    def dump_plugins(self) -> dict:
        plugins = {
            plugin.name: apischema.serialize(
                Optional[PluginResults],
                plugin.results,
            )
            for plugin in self.plugins
        }

        plugins["gateway"] = apischema.serialize(
            PluginResults,
            PluginResults(
                metadata=self.gateway_config,
            )
        )
        return plugins

    def dump_files(self, *, exclude_extensions: Optional[list[str]]) -> dict:
        exclude_extensions = exclude_extensions or []
        result = {}
        hash_to_contents = result["hash_to_contents"] = {}
        hash_to_file = result["hash_to_file"] = {}

        file_to_hash_dicts = [
            self.container.loaded_files,
            *[plugin.files_to_monitor for plugin in self.plugins],
            # TODO nested plugin results (pytmc)
        ]
        for file_to_hash in file_to_hash_dicts:
            for fn, hash_ in file_to_hash.items():
                if exclude_extensions:
                    if os.path.splitext(fn)[1].lower() in exclude_extensions:
                        continue

                if hash_ not in hash_to_contents:
                    hash_to_contents[hash_] = apischema.serialize(
                        self.get_file_info(fn, hash_)
                    )
                hash_to_file.setdefault(hash_, []).append(fn)

        return result

    async def dump_pv_relations(self, *, full: bool = True) -> dict:
        return apischema.serialize(
            PVRelationshipResponse(
                pv_relations=self.container.pv_relations,
                script_relations=self.script_relations,
                # can be generated on the client side easily with the cache
                ioc_to_pvs={},
            )
        )

    async def dump_pv_graphs(self, *, full: bool = True) -> dict:
        if not full:
            # No graphs at all in partial mode
            return {}

        relation_graphs = {}
        for record in self.container.database:
            if record in self.container.pv_relations:
                logger.info(
                    "Generating graph for record %s which has %d relation(s)",
                    record,
                    len(self.container.pv_relations[record]),
                )
                graph = self.get_graph(
                    (record,),
                    graph_type="record",
                    # format="svg",
                )
                # rendered = await self.get_graph_rendered(
                #     (record,),
                #     graph_type="record",
                #     format="svg",
                # )
                relation_graphs[record] = {
                    "source": graph.source,
                    # "svg": rendered.decode("utf-8"),
                }

        return relation_graphs

    def _prune_for_partial_dump(self) -> None:
        # TODO: this shouldn't have to be destructive
        logger.info("Pruning database definitions for a partial dump")
        for ioc in self.container.scripts.values():
            ioc.shell_state.database_definition = None

    async def dump(self, full: bool = True) -> dict:
        t0 = time.monotonic()

        dumped = {
            "plugins": self.dump_plugins(),
            "pv_graphs": await self.dump_pv_graphs(full=full),
            "pv_relations": await self.dump_pv_relations(full=full),
        }

        if not full:
            self._prune_for_partial_dump()

        dumped["iocs"] = self.dump_iocs()
        dumped["files"] = self.dump_files(
            exclude_extensions=[] if full else [".dbd"]
        )

        # Remove some duplicated information - PV relations are tracked
        # at the top-level:
        for item in dumped["iocs"].values():
            item["ioc"].pop("pv_relations")

        dumped["logs"] = self.get_log_messages()
        elapsed = time.monotonic() - t0
        logger.info(
            (
                "Dumped %d plugins, %d files, %d record link graphs, "
                "%d inter-IOC relationships, and %d IOCs in %.1f seconds"
            ),
            len(dumped["plugins"]),
            len(dumped["files"]),
            len(dumped["pv_graphs"]),
            len(dumped["pv_relations"]["script_relations"]),
            len(dumped["iocs"]),
            elapsed,
        )
        return dumped

    async def dump_to_file(
        self,
        filename: pathlib.Path,
        indent: Optional[int] = None,
        encoding: str = "utf-8",
        full: bool = True,
    ) -> None:
        if filename.suffix.lower() not in (".json", ".gz"):
            raise RuntimeError(
                f"Output format not supported yet: {filename.suffix}. "
                f"You probably want .json.gz, by the way, if you're intending "
                f"to use the frontend with these results."
            )

        output = await self.dump(full=full)
        dumped = json.dumps(output, sort_keys=True, indent=indent)
        if filename.suffix.lower() in (".json", ):
            with open(filename, "wt") as fp:
                print(dumped, file=fp)
        elif filename.suffix.lower() in (".gz", ):
            with gzip.open(filename, "wb") as fp:
                fp.write(dumped.encode(encoding))

    @property
    def iocs_by_name(self) -> Dict[str, IocMetadata]:
        """Dictionary of IOC name to IocMetadata."""
        return {
            ioc.name: ioc
            for ioc in self.ioc_metadata
        }

    def add_or_update_ioc_metadata(self, md: IocMetadata):
        """
        Add a new IOC to monitor by its metadata.

        Note: an assumption is made here that an IOC name is unique among all
        those loaded.
        """
        try:
            existing = self.iocs_by_name[md.name]
        except KeyError:
            self.ioc_metadata.append(md)
        else:
            existing.update(md)

    @property
    def update_count(self) -> int:
        """The number of times IOCs have been updated."""
        return self._update_count

    def get_log_messages(self) -> list[str]:
        """Get the server's log messages."""
        return list(
            _log_handler.messages
            if _log_handler is not None
            else ["Logger not initialized"]
        )

    async def stop(self):
        """Stop any background updates."""
        self.running = False
        await self.tasks.cancel_all(wait=True)

    async def async_init(self, app):
        self.running = True
        self.tasks.create(self._update_loop())
        logger.info(
            "Server plugins enabled: %s",
            ", ".join(plugin.name for plugin in self.plugins)
        )
        for plugin in self.plugins:
            self.tasks.create(self._update_plugin_loop(plugin))

    async def _update_plugin(self, plugin: ServerPluginSpec):
        """Update the specified plugin."""
        with common.time_context() as ctx:
            try:
                await plugin.update()
            except Exception:
                logger.exception(
                    "Failed to update plugin %r [%.1f s]",
                    plugin.name, ctx()
                )
            else:
                logger.info(
                    "Successfully updated plugin %r [%.1f s]",
                    plugin.name, ctx()
                )

    async def _update_plugin_loop(self, plugin: ServerPluginSpec):
        while self.running and self.update_count == 0 and plugin.after_iocs:
            # Wait until IOCs have been loaded before updating this one for the
            # first time.
            await asyncio.sleep(1)

        logger.info("Server plugin %r updates started.", plugin.name)

        while self.running:
            logger.info("Updating plugin: %s", plugin.name)
            await self._update_plugin(plugin)
            # for record, md in info["record_to_metadata"].items():
            #     ...
            await asyncio.sleep(settings.SERVER_SCAN_PERIOD)

        logger.info("Server plugin %r updates finished.", plugin.name)

    async def _update_script_step(self) -> bool:
        """Update scripts from the script loader."""
        logger.info("Checking for new or updated IOCs...")
        await self.update_script_loaders()

        logger.info("Checking for changed scripts and database files...")
        self._load_gateway_config()

        updated = self.get_updated_iocs()
        if not updated:
            logger.info("No changes found.")
            return False

        logger.info(
            "%d IOC%s changed.", len(updated),
            " has" if len(updated) == 1 else "s have"
        )
        for idx, ioc in enumerate(updated[:10], 1):
            logger.info("* %d: %s", idx, ioc.name)
        if len(updated) > 10:
            logger.info("... and %d more", len(updated) - 10)

        with common.time_context() as ctx:
            await self.update_iocs(updated)
            logger.info(
                "Updated %d IOCs in %.1f seconds", len(updated), ctx()
            )
        self.clear_cache()
        return True

    async def _update_loop(self):
        """Update scripts from the script loader and watch for updates."""
        while self.running:
            changes = await self._update_script_step()
            poll_period = settings.SERVER_SCAN_PERIOD
            if not changes:
                poll_period *= 2
            await asyncio.sleep(poll_period)

        logger.info("Server script updates finished.")

    async def update_script_loaders(self):
        """Update scripts from the script loader and watch for updates."""
        for loader in self.script_loaders:
            await loader.update()
            for _, md in loader.scripts.items():
                self.add_or_update_ioc_metadata(md)

    def get_updated_iocs(self) -> List[IocMetadata]:
        """Check loaded IOCs for any changes."""
        updated = [
            md for md in self.ioc_metadata
            if not md.is_up_to_date()
        ]
        for item in list(updated):
            if not item.script or not item.script.exists() or item.looks_like_sh:
                if self.update_count == 0:
                    # Don't attempt another load unless the file exists
                    updated.remove(item)

        return updated

    def _replace_metadata(
        self,
        old_md: IocMetadata,
        new_md: IocMetadata
    ) -> int:
        """Replace the provided metadata information with a new one."""
        idx = self.ioc_metadata.index(old_md)
        self.ioc_metadata.remove(old_md)
        self.ioc_metadata.insert(idx, new_md)
        return idx

    async def update_iocs(self, iocs: List[IocMetadata]):
        """Reload the provided IOCs."""
        async for md, loaded in load_startup_scripts_with_metadata(
            *iocs, standin_directories=self.standin_directories
        ):
            self.container.add_loaded_ioc(loaded)
            self._replace_metadata(md, loaded.metadata)

            # Let plugins update, if possible
            await asyncio.sleep(0)

        with common.time_context() as ctx:
            self.script_relations = graph.build_script_relations(
                self.container.database,
                self.container.pv_relations,
            )

        logger.info("Updated script relations in %.1f s", ctx())
        self._update_count += 1

    def get_plugin_info(self, allow_list: Optional[List[str]] = None) -> Dict[str, Any]:
        """Get plugin information as a dictionary."""
        if allow_list is None:
            allow_list = [plugin.name for plugin in self.plugins]

        return {
            plugin.name: plugin.results_json
            for plugin in self.plugins
            if plugin.name in allow_list and plugin.results
        }

    def get_plugin_nested_keys(self, plugin_name: str) -> List[str]:
        """Get plugin custom nested metadata keys."""
        plugin = self.plugins_by_name[plugin_name]
        if plugin.results and plugin.results.nested:
            return list(plugin.results.nested)
        return []

    def get_plugin_nested_info(self, plugin_name: str, key: str) -> Any:
        """Get plugin custom nested metadata info."""
        results = self.plugins_by_name[plugin_name].results
        return results.nested[key] if results else {}

    def _load_gateway_config(self):
        if not self.gateway_config_path:
            logger.warning(
                "Gateway path not set; gateway configuration will not be loaded"
            )
            return

        if self.gateway_config is None:
            logger.info("Loading gateway configuration for the first time...")
            self.gateway_config = gateway.GatewayConfig(
                self.gateway_config_path
            )
        else:
            logger.info("Updating gateway configuration...")
            self.gateway_config.update_changed()

        for filename, pvlist in self.gateway_config.pvlists.items():
            if pvlist.hash is not None:
                logger.debug("New gateway file: %s (%s)", filename, pvlist.hash)
                self.container.loaded_files[str(filename)] = pvlist.hash

    def get_record_annnotations(
        self,
        ioc: LoadedIoc,
        instance: RecordInstance,
    ) -> Dict[StringWithContext, Any]:
        """
        Get IOC/plugin-provided annotations for the given record.
        """
        md = {}

        if not instance.is_pva:
            # For now, V3 only
            gateway_key = StringWithContext("gateway", context=())
            md[gateway_key] = apischema.serialize(
                self.get_gateway_matches(instance.name)
            )

            md.update(ioc.shell_state.annotate_record(instance))

        for plugin in self.plugins:
            if not plugin.results:
                continue

            info = list(plugin.results.find_record_metadata(instance.name))
            if info:
                plugin_key = StringWithContext(plugin.name, context=())
                instance.metadata[plugin_key] = info

        return md

    def annotate_whatrec(self, ioc: LoadedIoc, what: WhatRecord) -> WhatRecord:
        """
        Annotate WhatRecord instances with things ServerState knows about.
        """
        matches = [
            (what.record.instance if what.record else None),
            what.pva_group
        ]
        for instance in matches:
            if instance is None:
                continue

            md = self.get_record_annnotations(ioc, instance)
            instance.metadata.update(md)

        return what

    def whatrec(self, pvname: str) -> List[WhatRecord]:
        """Find WhatRecord matches."""
        results = []
        for loaded_ioc in self.container.scripts.values():
            what = loaded_ioc.whatrec(pvname)
            if what is not None:
                self.annotate_whatrec(loaded_ioc, what)
                results.append(what)
        return results

    @property
    def aliases(self) -> Dict[str, str]:
        """The CA/V3 aliases."""
        return self.container.aliases

    @property
    def database(self) -> Dict[str, RecordInstance]:
        """The CA/V3 Database of records."""
        return self.container.database

    @property
    def pva_database(self) -> Dict[str, RecordInstance]:
        """The pvAccess Database of groups/records."""
        return self.container.pva_database

    def get_graph(self, pv_names: Tuple[str, ...], graph_type: str) -> graphviz.Digraph:
        if graph_type == "record":
            return self.get_link_graph(tuple(pv_names))
        if graph_type == "script":
            return self.get_script_graph(tuple(pv_names))
        raise RuntimeError("Invalid graph type")

    def get_script_graph(self, pv_names: Tuple[str, ...]) -> graphviz.Digraph:
        if len(pv_names) > settings.MAX_RECORDS:
            raise TooManyRecordsError()

        if not pv_names:
            return graphviz.Digraph()
        gr = graph.graph_script_relations(
            database=self.database,
            limit_to_records=pv_names,
            script_relations=self.script_relations,
        )
        return gr.to_digraph(font_name="Courier")

    def get_ioc_to_pvs(self, pv_names: Tuple[str, ...]) -> Dict[str, List[str]]:
        ioc_to_pvs = {}
        for pv in pv_names:
            try:
                owner = self.container.database[pv].owner or "unknown"
            except KeyError:
                owner = "unknown"

            if owner not in ioc_to_pvs:
                ioc_to_pvs[owner] = []
            ioc_to_pvs[owner].append(pv)
        return ioc_to_pvs

    def get_pv_relations(
        self,
        pv_names: Tuple[str, ...],
        *,
        full: bool = False,
    ) -> Union[PVRelationshipResponse, PVShortRelationshipResponse]:
        # TODO: pv_names
        if full:
            return PVRelationshipResponse(
                pv_relations=self.container.pv_relations,
                script_relations=self.script_relations,
                ioc_to_pvs=self.get_ioc_to_pvs(tuple(self.container.pv_relations))
            )
        return PVShortRelationshipResponse.from_pv_relations(
            pv_relations=self.container.pv_relations,
            script_relations=self.script_relations,
            ioc_to_pvs=self.get_ioc_to_pvs(tuple(self.container.pv_relations))
        )

    def get_link_graph(self, pv_names: Tuple[str, ...]) -> graphviz.Digraph:
        if len(pv_names) > settings.MAX_RECORDS:
            raise TooManyRecordsError()

        if not pv_names:
            return graphviz.Digraph()
        gr = graph.graph_links(
            database=self.database,
            starting_records=list(pv_names),
            sort_fields=True,
            relations=self.container.pv_relations,
        )
        digraph = gr.to_digraph(font_name="Courier")
        return digraph

    def clear_cache(self):
        for method in [
            # self.get_graph,
            # self.get_graph_rendered,
            self.get_gateway_matches,
            self.get_matching_pvs,
            self.get_matching_iocs,
            self.script_info_from_loaded_file,
            self.get_duplicates,
        ]:
            method.cache_clear()

    def is_loaded_file(self, fn) -> bool:
        """Is ``fn`` a file that was loaded?"""
        fn = str(fn)
        if fn in self.container.loaded_files:
            return True

        return any(
            plugin.results.is_loaded_file(fn)
            for plugin in self.plugins
            if plugin.results is not None
        )

    @functools.lru_cache(maxsize=2048)
    def script_info_from_loaded_file(self, fn) -> common.IocshScript:
        with open(fn, "rt") as fp:
            lines = fp.read().splitlines()

        result = []
        for lineno, line in enumerate(lines, 1):
            result.append(
                common.IocshResult(
                    context=(LoadContext(fn, lineno),),
                    line=line,
                )
            )
        return common.IocshScript(path=fn, lines=tuple(result))

    @functools.lru_cache(maxsize=2048)
    def get_gateway_matches(self, pvname: str) -> Optional[gateway.PVListMatches]:
        """Get gateway matches for the given pvname."""
        if self.gateway_config is None:
            return None
        return self.gateway_config.get_matches(pvname)

    @functools.lru_cache(maxsize=20)
    def get_duplicates(
        self, patterns: Tuple[str, ...], use_regex: bool = False
    ) -> Dict[str, List[str]]:
        """Get duplicate PVs from the matching IOC(s)."""
        iocs = self.get_matching_iocs(patterns, use_regex=use_regex)
        seen = collections.defaultdict(list)
        for ioc in iocs:
            shell_state: ShellState = ioc.shell_state
            for record in set(shell_state.database).union(shell_state.pva_database):
                seen[record].append(ioc.name)
        return {
            record: iocs
            for record, iocs in seen.items()
            if len(iocs) > 1
        }

    @functools.lru_cache(maxsize=2048)
    def get_matching_pvs(self, patterns: Tuple[str, ...], use_regex: bool = False) -> List[str]:
        """
        Get matching PV names given pattern(s).

        Parameters
        ----------
        patterns : list of str
            List of patterns in glob or regex format.

        use_regex : bool, optional
            Interpret patterns as glob (False) or regex (True).
        """
        try:
            regex = compile_patterns(patterns, use_regex=use_regex)
        except re.error:
            return []

        pv_names = set(self.database) | set(self.pva_database) | set(self.aliases)
        return [pv_name for pv_name in sorted(pv_names) if regex.match(pv_name)]

    @functools.lru_cache(maxsize=2048)
    def get_matching_iocs(
        self, patterns: Tuple[str, ...], use_regex: bool = False
    ) -> List[LoadedIoc]:
        """
        Get matching IOCs given pattern(s).

        Parameters
        ----------
        patterns : list of str
            List of patterns in glob or regex format.

        use_regex : bool, optional
            Interpret patterns as glob (False) or regex (True).
        """
        try:
            regex = compile_patterns(patterns, use_regex=use_regex)
        except re.error:
            return []

        def by_name(ioc: LoadedIoc):
            return ioc.name

        return [
            loaded_ioc
            for loaded_ioc in sorted(self.container.scripts.values(), key=by_name)
            if regex.match(loaded_ioc.script.path)
            or regex.match(loaded_ioc.metadata.name)
        ]

    async def get_graph_rendered(
        self, pv_names: Tuple[str, ...], format: str, graph_type: str
    ) -> bytes:
        """
        Get a rendered PV relationship graph of the provided PVs.

        Parameters
        ----------
        pv_names : tuple of str
            PV names.

        format : str
            The format of the graph (e.g., pdf, png).

        graph_type : { "record", "script" }
            The type of graph to generate.
        """
        graph = self.get_graph(pv_names, graph_type=graph_type)

        with tempfile.NamedTemporaryFile(suffix=f".{format}") as source_file:
            rendered_filename = await graph.async_render(
                source_file.name, format=format
            )

        with open(rendered_filename, "rb") as fp:
            return fp.read()


class ServerHandler:
    routes = web.RouteTableDef()

    def __init__(self, state: ServerState):
        self.state = state

    async def async_init(self, app):
        await self.state.async_init(app)

    @routes.get("/api/pv/info")
    async def api_pv_get_info(self, request: web.Request):
        pv_names = request.query.getall("pv")
        info = {pv_name: self.state.whatrec(pv_name) for pv_name in pv_names}
        " {pv_name: ..., present: True, info=[whatrec]}"
        return serialized_response(
            {
                pv_name: PVGetInfo(
                    pv_name=pv_name,
                    present=(pv_name in self.state.database or
                             pv_name in self.state.pva_database),
                    info=[obj for obj in info[pv_name]],
                )
                for pv_name in pv_names
            }
        )

    @routes.get("/api/pv/matches")
    async def api_pv_get_matches(self, request: web.Request):
        max_matches = int(request.query.get("max", "200"))
        use_regex, patterns = get_patterns(request.query)
        matches = self.state.get_matching_pvs(patterns, use_regex=use_regex)
        if max_matches > 0:
            matches = matches[:max_matches]

        return serialized_response(
            PVGetMatchesResponse(
                patterns=patterns,
                regex=use_regex,
                matches=matches,
            )
        )

    @routes.get("/api/ioc/matches")
    async def api_ioc_get_matches(self, request: web.Request):
        # Ignore max for now. This is not much in the way of information.
        # max_matches = int(request.query.get("max", "1000"))
        use_regex, patterns = get_patterns(request.query)
        match_metadata = [
            loaded_ioc.metadata
            for loaded_ioc in self.state.get_matching_iocs(patterns, use_regex=use_regex)
        ]
        return serialized_response(
            IocGetMatchesResponse(
                patterns=patterns,
                regex=use_regex,
                matches=match_metadata,
            )
        )

    @routes.get("/api/ioc/pvs")
    async def api_ioc_get_pvs(self, request: web.Request):
        use_regex, ioc_patterns = get_patterns(request.query, key="ioc")
        _, record_patterns = get_patterns(request.query, key="pv")
        response = IocGetMatchingRecordsResponse(
            ioc_patterns=ioc_patterns,
            record_patterns=record_patterns,
            regex=use_regex,
            matches=[],
        )

        def get_all_records(shell_state):
            yield from shell_state.database.items()
            yield from shell_state.pva_database.items()

        try:
            pv_glob_re = compile_patterns(
                record_patterns,
                use_regex=use_regex,
            )
        except re.error:
            raise web.HTTPBadRequest()

        for loaded_ioc in self.state.get_matching_iocs(
            ioc_patterns, use_regex=response.regex,
        ):
            record_matches = [
                rec_info.to_summary()
                for rec, rec_info in sorted(get_all_records(loaded_ioc.shell_state))
                if pv_glob_re.match(rec)
            ]
            if record_matches:
                response.matches.append((loaded_ioc.metadata, record_matches))

        return serialized_response(response)

    @routes.get("/api/pv/duplicates")
    async def api_pv_get_duplicates(self, request: web.Request):
        """Get record names duplicated among two or more IOCs."""
        use_regex, patterns = get_patterns(request.query)
        return serialized_response(
            IocGetDuplicatesResponse(
                patterns=patterns,
                regex=use_regex,
                duplicates=self.state.get_duplicates(patterns, use_regex=use_regex),
            )
        )

    # @routes.get("/api/graphql/query")
    # async def api_graphql_query(self, request: web.Request):
    # TODO: ...

    @routes.get("/api/plugin/info")
    async def api_plugin_info(self, request: web.Request):
        plugins = request.query.get("plugin", "all")
        allow_list = None if plugins == "all" else plugins.split(" ")
        return serialized_response(self.state.get_plugin_info(allow_list))

    @routes.get("/api/plugin/nested/keys")
    async def api_plugin_nested_keys(self, request: web.Request):
        try:
            plugin = request.query["plugin"]
            keys = self.state.get_plugin_nested_keys(plugin)
        except KeyError:
            raise web.HTTPBadRequest()
        return serialized_response(keys)

    @routes.get("/api/plugin/nested/info")
    async def api_plugin_nested_info(self, request: web.Request):
        try:
            plugin = request.query["plugin"]
            key = request.query["key"]
            info = self.state.get_plugin_nested_info(plugin, key)
        except KeyError:
            raise web.HTTPBadRequest()
        return serialized_response(info)

    @routes.get("/api/gateway/info")
    async def api_gateway_info(self, request: web.Request):
        return serialized_response(self.state.gateway_config)

    @routes.get("/api/file/info")
    async def api_ioc_info(self, request: web.Request):
        # TODO: support new 'raw' response
        # script_name = pathlib.Path(request.query["file"])
        filename = request.query["file"]
        ioc_name = self.state.container.startup_script_to_ioc.get(filename, None)
        if ioc_name:
            loaded_ioc = self.state.container.scripts[ioc_name]
            script_info = loaded_ioc.script
            ioc_md = loaded_ioc.metadata
        else:
            # Making this dual-purpose: script, db, or any loaded file
            ioc_md = None
            if not self.state.is_loaded_file(filename):
                raise web.HTTPBadRequest()

            script_info = self.state.script_info_from_loaded_file(filename)

        return serialized_response(
            {
                "script": script_info,
                "ioc": ioc_md,
            }
        )

    async def get_graph(self, pv_names: List[str], use_glob: bool = False,
                        graph_type: str = "record",
                        format: str = "pdf"):
        if use_glob:
            pv_names = self.state.get_matching_pvs(pv_names)

        if format == "dot":
            try:
                digraph = self.state.get_graph(
                    tuple(pv_names),
                    graph_type=graph_type
                )
            except TooManyRecordsError as ex:
                raise web.HTTPBadRequest() from ex

            return web.Response(
                content_type="text/vnd.graphviz",
                body=digraph.source,
            )

        rendered = await self.state.get_graph_rendered(
            tuple(pv_names), format=format, graph_type=graph_type
        )
        try:
            content_type = {
                "pdf": "application/pdf",
                "png": "image/png",
                "svg": "image/svg+xml",
            }[format]
        except KeyError as ex:
            raise web.HTTPBadRequest() from ex

        return web.Response(
            content_type=content_type,
            body=rendered,
        )

    @routes.get("/api/logs/get")
    async def api_logs_get(self, request: web.Request):
        return web.json_response(self.state.get_log_messages())

    @routes.get("/api/pv/relations")
    async def api_pv_get_relations(self, request: web.Request):
        use_regex, pv_names = get_patterns(request.query, key="pv")
        pv_names = self.state.get_matching_pvs(pv_names, use_regex=use_regex)
        full = request.query.get("full", "false") in TRUE_VALUES
        return serialized_response(
            self.state.get_pv_relations(pv_names=pv_names, full=full)
        )

    @routes.get("/api/pv/graph")
    async def api_pv_get_record_graph(self, request: web.Request):
        return await self.get_graph(
            pv_names=request.query.getall("pv"),
            use_glob=request.query.get("glob", "false") in TRUE_VALUES,
            format=request.query["format"],
            graph_type="record",
        )

    @routes.get("/api/pv/script-graph")
    async def api_pv_get_script_graph(self, request: web.Request):
        return await self.get_graph(
            pv_names=request.query.getall("pv"),
            use_glob=request.query.get("glob", "false") in TRUE_VALUES,
            format=request.query["format"],
            graph_type="script",
        )


def add_routes(app: web.Application, handler: ServerHandler):
    routes = web.RouteTableDef()
    for route in handler.routes:
        if isinstance(route, web.StaticDef):
            routes.static(route.prefix, route.path, **route.kwargs)
        elif isinstance(route, web.RouteDef):
            method = getattr(handler, route.handler.__name__)
            routes.route(route.method, route.path, **route.kwargs)(method)
    app.router.add_routes(routes)
    return routes


def configure_logging(loggers=None):
    global _log_handler
    _log_handler = ServerLogHandler()

    loggers = loggers or ["whatrecord"]
    for logger_name in loggers:
        logging.getLogger(logger_name).addHandler(_log_handler)


def _new_server_state(
    scripts: Optional[List[str]] = None,
    script_loader: Optional[List[str]] = None,
    archive_management_url: Optional[str] = None,
    gateway_config: Optional[str] = None,
    standin_directory: Optional[Union[List, Dict]] = None,
) -> ServerState:
    """New ServerState from command-line arguments."""
    scripts = scripts or []
    script_loader = script_loader or []

    standin_directory = standin_directory or {}
    if not isinstance(standin_directory, dict):
        standin_directory = dict(path.split("=", 1) for path in standin_directory)

    plugins = []
    if "happi" in settings.PLUGINS:
        plugins.append(
            ServerPluginSpec(
                name="happi",
                module="whatrecord.plugins.happi",
                executable=None,
            )
        )

    if "twincat_pytmc" in settings.PLUGINS:
        plugins.append(
            ServerPluginSpec(
                name="twincat_pytmc",
                module="whatrecord.plugins.twincat_pytmc",
                executable=None,
                after_iocs=True,
            )
        )

    if "netconfig" in settings.PLUGINS:
        plugins.append(
            ServerPluginSpec(
                name="netconfig",
                module="whatrecord.plugins.netconfig",
                executable=None,
                after_iocs=False,
            )
        )

    if "epicsarch" in settings.PLUGINS:
        plugins.append(
            ServerPluginSpec(
                name="epicsarch",
                module="whatrecord.plugins.epicsarch",
                executable=None,
                after_iocs=False,
            )
        )

    return ServerState(
        startup_scripts=scripts,
        script_loaders=script_loader,
        standin_directories=standin_directory,
        gateway_config=gateway_config,
        plugins=plugins,
    )


def _new_server(
    handler: ServerHandler,
    port: int,
    run: bool = False,
) -> web.Application:
    """Create a new aiohttp Application for the given ServerHandler."""
    app = web.Application()
    add_routes(app, handler)

    # Set the environment variable for plugins or subprocesses to be able to
    # query the server.
    os.environ["WHATRECORD_SERVER"] = f"http://localhost:{port}"

    app.on_startup.append(handler.async_init)
    if run:
        try:
            web.run_app(app, port=port)
        except KeyboardInterrupt:
            ...
    return app


def main(
    scripts: Optional[List[str]] = None,
    script_loader: Optional[List[str]] = None,
    archive_management_url: Optional[str] = None,
    gateway_config: Optional[str] = None,
    standin_directory: Optional[Union[List, Dict]] = None,
    port: int = 8898,
    use_tracemalloc: bool = False,
    offline_dump_target: Optional[AnyPath] = None,
    start: bool = True,
    partial_dump: bool = False,
):
    if use_tracemalloc and tracemalloc is not None:
        tracemalloc.start()

    configure_logging()

    state: ServerState = _new_server_state(
        scripts=scripts,
        script_loader=script_loader,
        archive_management_url=archive_management_url,
        gateway_config=gateway_config,
        standin_directory=standin_directory,
    )
    handler = ServerHandler(state)
    if offline_dump_target is None:
        # Run the server.
        run = True
    else:
        # Dump to a file - but don't start the server.
        asyncio.run(state.trigger_manual_update(annotate_records=True))
        asyncio.run(
            state.dump_to_file(pathlib.Path(offline_dump_target), full=not partial_dump)
        )
        run = False

    app = _new_server(handler, port=port, run=run)

    if use_tracemalloc and tracemalloc is not None:
        global tracemalloc_snapshot
        tracemalloc_snapshot = tracemalloc.take_snapshot()

    return app, handler
